/*
 * $Id: hexbincnt.prg 2961 2021-03-14 10:25:21Z df7be $
 *
 * HWGUI - Harbour Win32 and Linux (GTK) GUI library
 *
 *  hexbincnt.prg
 *
 * Sample program for usage of image resources from
 * Hex definitions.
 *
 * With this system the image resources are
 * compiled in the exe of the HWGUI program.
 * For multi platform purposes,
 * a good substitute for Windows
 * resource compiler, it is "Windows only".
 *
 * Copyright 2020 Wilfried Brunken, DF7BE 
 * https://sourceforge.net/projects/cllog/
 */
 
    * Status:
    *  WinAPI   :  Yes
    *  GTK/Linux:  Yes
    *  GTK/Win  :  Yes

* --------- Instructions -------------
*
* 1.) Compile the Utility "file2hex.prg".
*     Path is "utils\bincnt".
*
* 2.) Start file2hex.exe
*     - Select desired image in the file selection dialog
*     - Wait for message of creating hex file
*
* 3.) Open generated file "hexdump.txt" with 
*     your preferred text editor, copy and
*     paste the variable definition into
*     your program code file.
*
* 4.) Delete file hexdump.txt, if not used any more.
*
* Repeat this for every image file.
* Every run of "file2hex" overwrites hexdump.txt.
* 
* You need a 2 variables for every image
*  1.) cHex.. : contains the Hex string (or set as a "long" parameter to hwg_cHex2Bin() )
*  2.) cVal.. : contains the binary value of the image.
* For every image the following function calls necessary:
*  - cVal.. := hwg_cHex2Bin(cHex..)
*  - oImage := HBitmap():AddString( "<resname>", cVal.. )
* Now the resources are ready to use.
*
* Don't forget, that only *.BMP and *.ICO formats are supported in WinAPI,
* so we strictly recommend to convert your images into one of this both
* formats for multi platform applications.

#include "hwgui.ch"

MEMVAR cHexAstro , cHexok , cHexopen , cHexexit , cHexdoor , cHexnext    && Hex dumps 
MEMVAR cValAstro , cValok , cValopen , cValexit , cValdoor , cValnext    && Resource contents

FUNCTION Main

LOCAL cImageDir, cppath , oIcon, oBitmap , oToolbar , oFileOpen , oQuit , oMainW , oFontMain
LOCAL htab, nbut , oBMPExit , oPNGDoor , oBtnDoor , ojpeg , oBtnjpeg , oastropng
LOCAL cDirSep := hwg_GetDirSep()

PUBLIC cHexAstro , cHexok , cHexopen , cHexexit , cHexdoor , cHexnext    && Hex dumps 
PUBLIC cValAstro , cValok , cValopen , cValexit , cValdoor , cValnext    && Resource contents

htab := 0
nbut := 0

* Fill variables with hex values
Init_Hexvars()

* Convert them all to binary.
cValAstro := hwg_cHex2Bin ( cHexAstro )
cValok    := hwg_cHex2Bin ( cHexok )
cValopen  := hwg_cHex2Bin ( cHexopen )
cValexit  := hwg_cHex2Bin ( cHexexit )
cValdoor  := hwg_cHex2Bin ( cHexdoor )
cValnext  := hwg_cHex2Bin ( cHexnext )

* Load contents from hex resources into image objects.
oIcon     := HIcon():AddString( "ok" , cValok )
oBitmap   := HBitmap():AddString( "open" , cValopen )
oBMPExit  := HBitmap():AddString( "exit" , cValexit )
oPNGDoor  := HBitmap():AddString( "door" , cValdoor )
ojpeg     := HBitmap():AddString( "next" , cValnext )
oastropng := HBitmap():AddString( "astro", cValAstro )

#ifdef __PLATFORM__WINDOWS
   PREPARE FONT oFontMain NAME "MS Sans Serif" WIDTH 0 HEIGHT -14
#else
   PREPARE FONT oFontMain NAME "Sans" WIDTH 0 HEIGHT 12 
#endif

INIT WINDOW oMainW  ;
   FONT oFontMain  ;
   TITLE "Bitmap Hex container sample" AT 0,0 SIZE 300 , 200 ;
   ICON oIcon STYLE WS_POPUP +  WS_CAPTION + WS_SYSMENU ;
 
   
* @ 0, 0 TOOLBAR oToolbar OF oMainW SIZE  299 , 50 
 @ 0,0 PANEL oToolbar OF oMainW SIZE 300 , 50 ON SIZE ANCHOR_TOPABS + ANCHOR_LEFTABS + ANCHOR_RIGHTABS 


@ htab+(nbut*32), 3 OWNERBUTTON oFileOpen OF oToolbar ;
   ON CLICK { | | FileOpen()} ;
   SIZE 28,24 FLAT ;
   BITMAP oBitmap ;
   TRANSPARENT COLOR hwg_ColorC2N("#DCDAD5") COORDINATES 0,4,0,0 ; 
   TOOLTIP "Open File"
   
   nbut += 1

@ htab+(nbut*32),3 OWNERBUTTON oQuit OF oToolbar ;
   ON CLICK { | | oMainW:Close()} ;
   SIZE 28,24 FLAT ;
   BITMAP oBMPExit ; 
   TRANSPARENT COLOR hwg_ColorC2N("#DCDAD5") COORDINATES 0,4,0,0 ; 
   TOOLTIP "Terminate Program"

   
   
   nbut += 1

* !!!!! PNG not supported   
@ htab+(nbut*32),3 OWNERBUTTON oBtnDoor OF oToolbar ;
   ON CLICK { | | OpenDoor()} ;
   SIZE 28,24 FLAT ;
   BITMAP oPNGDoor ; 
   TRANSPARENT COLOR hwg_ColorC2N("#DCDAD5") COORDINATES 0,4,0,0 ; 
   TOOLTIP "Open the door"
   
  nbut += 1 

@ htab+(nbut*32),3 OWNERBUTTON oBtnjpeg OF oToolbar ;
   ON CLICK { | | ClickJpeg()} ;
   SIZE 28,24 FLAT ;
   BITMAP ojpeg ; 
   TRANSPARENT COLOR hwg_ColorC2N("#DCDAD5") COORDINATES 0,4,0,0 ; 
   TOOLTIP "JPEG image" 
  

 @ 60 , 100 SAY "astro.png" SIZE 100, 20
 @ 60 , 100 BITMAP oastropng 
 
   oMainW:Activate()
RETURN NIL

FUNCTION FileOpen
 hwg_msginfo("You have clicked >FileOpen<")
RETURN NIL

FUNCTION OpenDoor
 hwg_msginfo("You have clicked >Open the door<")
RETURN NIL

FUNCTION ClickJpeg
 hwg_msginfo("You have clicked >JPEG image<")
RETURN NIL


* ==============================================
* Init hex defintions of binary resources
* Created with utility "file2hex.prg" from image file.
* ==============================================
FUNCTION Init_Hexvars

* --------- ok.ico ----------------
cHexok  := ;
"00 00 01 00 01 00 20 20 00 00 00 00 00 00 A8 08 " + ;
"00 00 16 00 00 00 28 00 00 00 20 00 00 00 40 00 " + ;
"00 00 01 00 08 00 00 00 00 00 00 04 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 80 00 00 80 00 00 00 80 80 00 80 00 " + ;
"00 00 80 00 80 00 80 80 00 00 C0 C0 C0 00 C0 DC " + ;
"C0 00 F0 CA A6 00 04 04 04 00 08 08 08 00 0C 0C " + ;
"0C 00 11 11 11 00 16 16 16 00 1C 1C 1C 00 22 22 " + ;
"22 00 29 29 29 00 55 55 55 00 4D 4D 4D 00 42 42 " + ;
"42 00 39 39 39 00 80 7C FF 00 50 50 FF 00 93 00 " + ;
"D6 00 FF EC CC 00 C6 D6 EF 00 D6 E7 E7 00 90 A9 " + ;
"AD 00 00 00 33 00 00 00 66 00 00 00 99 00 00 00 " + ;
"CC 00 00 33 00 00 00 33 33 00 00 33 66 00 00 33 " + ;
"99 00 00 33 CC 00 00 33 FF 00 00 66 00 00 00 66 " + ;
"33 00 00 66 66 00 00 66 99 00 00 66 CC 00 00 66 " + ;
"FF 00 00 99 00 00 00 99 33 00 00 99 66 00 00 99 " + ;
"99 00 00 99 CC 00 00 99 FF 00 00 CC 00 00 00 CC " + ;
"33 00 00 CC 66 00 00 CC 99 00 00 CC CC 00 00 CC " + ;
"FF 00 00 FF 66 00 00 FF 99 00 00 FF CC 00 33 00 " + ;
"00 00 33 00 33 00 33 00 66 00 33 00 99 00 33 00 " + ;
"CC 00 33 00 FF 00 33 33 00 00 33 33 33 00 33 33 " + ;
"66 00 33 33 99 00 33 33 CC 00 33 33 FF 00 33 66 " + ;
"00 00 33 66 33 00 33 66 66 00 33 66 99 00 33 66 " + ;
"CC 00 33 66 FF 00 33 99 00 00 33 99 33 00 33 99 " + ;
"66 00 33 99 99 00 33 99 CC 00 33 99 FF 00 33 CC " + ;
"00 00 33 CC 33 00 33 CC 66 00 33 CC 99 00 33 CC " + ;
"CC 00 33 CC FF 00 33 FF 33 00 33 FF 66 00 33 FF " + ;
"99 00 33 FF CC 00 33 FF FF 00 66 00 00 00 66 00 " + ;
"33 00 66 00 66 00 66 00 99 00 66 00 CC 00 66 00 " + ;
"FF 00 66 33 00 00 66 33 33 00 66 33 66 00 66 33 " + ;
"99 00 66 33 CC 00 66 33 FF 00 66 66 00 00 66 66 " + ;
"33 00 66 66 66 00 66 66 99 00 66 66 CC 00 66 99 " + ;
"00 00 66 99 33 00 66 99 66 00 66 99 99 00 66 99 " + ;
"CC 00 66 99 FF 00 66 CC 00 00 66 CC 33 00 66 CC " + ;
"99 00 66 CC CC 00 66 CC FF 00 66 FF 00 00 66 FF " + ;
"33 00 66 FF 99 00 66 FF CC 00 CC 00 FF 00 FF 00 " + ;
"CC 00 99 99 00 00 99 33 99 00 99 00 99 00 99 00 " + ;
"CC 00 99 00 00 00 99 33 33 00 99 00 66 00 99 33 " + ;
"CC 00 99 00 FF 00 99 66 00 00 99 66 33 00 99 33 " + ;
"66 00 99 66 99 00 99 66 CC 00 99 33 FF 00 99 99 " + ;
"33 00 99 99 66 00 99 99 99 00 99 99 CC 00 99 99 " + ;
"FF 00 99 CC 00 00 99 CC 33 00 66 CC 66 00 99 CC " + ;
"99 00 99 CC CC 00 99 CC FF 00 99 FF 00 00 99 FF " + ;
"33 00 99 CC 66 00 99 FF 99 00 99 FF CC 00 99 FF " + ;
"FF 00 CC 00 00 00 99 00 33 00 CC 00 66 00 CC 00 " + ;
"99 00 CC 00 CC 00 99 33 00 00 CC 33 33 00 CC 33 " + ;
"66 00 CC 33 99 00 CC 33 CC 00 CC 33 FF 00 CC 66 " + ;
"00 00 CC 66 33 00 99 66 66 00 CC 66 99 00 CC 66 " + ;
"CC 00 99 66 FF 00 CC 99 00 00 CC 99 33 00 CC 99 " + ;
"66 00 CC 99 99 00 CC 99 CC 00 CC 99 FF 00 CC CC " + ;
"00 00 CC CC 33 00 CC CC 66 00 CC CC 99 00 CC CC " + ;
"CC 00 CC CC FF 00 CC FF 00 00 CC FF 33 00 99 FF " + ;
"66 00 CC FF 99 00 CC FF CC 00 CC FF FF 00 CC 00 " + ;
"33 00 FF 00 66 00 FF 00 99 00 CC 33 00 00 FF 33 " + ;
"33 00 FF 33 66 00 FF 33 99 00 FF 33 CC 00 FF 33 " + ;
"FF 00 FF 66 00 00 FF 66 33 00 CC 66 66 00 FF 66 " + ;
"99 00 FF 66 CC 00 CC 66 FF 00 FF 99 00 00 FF 99 " + ;
"33 00 FF 99 66 00 FF 99 99 00 FF 99 CC 00 FF 99 " + ;
"FF 00 FF CC 00 00 FF CC 33 00 FF CC 66 00 FF CC " + ;
"99 00 FF CC CC 00 FF CC FF 00 FF FF 33 00 CC FF " + ;
"66 00 FF FF 99 00 FF FF CC 00 66 66 FF 00 66 FF " + ;
"66 00 66 FF FF 00 FF 66 66 00 FF 66 FF 00 FF FF " + ;
"66 00 21 00 A5 00 5F 5F 5F 00 77 77 77 00 86 86 " + ;
"86 00 96 96 96 00 CB CB CB 00 B2 B2 B2 00 D7 D7 " + ;
"D7 00 DD DD DD 00 E3 E3 E3 00 EA EA EA 00 F1 F1 " + ;
"F1 00 F8 F8 F8 00 F0 FB FF 00 A4 A0 A0 00 80 80 " + ;
"80 00 00 00 FF 00 00 FF 00 00 00 FF FF 00 FF 00 " + ;
"00 00 FF 00 FF 00 FF FF 00 00 FF FF FF 00 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A FF FF FF 6D EB 6D EB " + ;
"6D EB 6D EB 6D EB 6D 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A FF 33 02 FF EB 6D EB " + ;
"6D EB 6D EB 6D EB 6D 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A FF 33 33 02 02 FF 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A FF 33 33 02 02 FF 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A FF 33 33 33 33 02 02 FF 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A FF 33 33 33 33 02 02 FF 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A FF 33 33 BC E4 33 33 02 02 FF " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A FF 33 33 BC E4 33 33 02 02 FF " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A FF 33 33 BC E4 E4 FF 33 33 02 02 " + ;
"FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A FF E4 E4 E4 BC FF 0A FF 33 02 02 " + ;
"FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A FF E4 E4 E4 FF 0A FF 33 33 02 " + ;
"02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A FF FF FF 0A 0A 0A FF 33 02 " + ;
"02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF 33 33 " + ;
"02 02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF 33 " + ;
"02 02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF 33 " + ;
"33 02 02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF " + ;
"33 02 02 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF " + ;
"33 33 BC E4 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"FF BC E4 E4 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"FF E4 E4 FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A FF FF 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A " + ;
"0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A 0A FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF E0 00 7F FF E0 00 7F FF C0 " + ;
"FF FF FF C0 FF FF FF 80 7F FF FF 80 7F FF FF 00 " + ;
"3F FF FF 00 3F FF FE 00 1F FF FE 04 1F FF FF 04 " + ;
"0F FF FF 8E 0F FF FF FE 07 FF FF FF 07 FF FF FF " + ;
"03 FF FF FF 83 FF FF FF 81 FF FF FF C1 FF FF FF " + ;
"C3 FF FF FF E7 FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF "

* --------- open.bmp ----------------
cHexopen  := ;
"42 4D F6 00 00 00 00 00 00 00 76 00 00 00 28 00 " + ;
"00 00 10 00 00 00 10 00 00 00 01 00 04 00 00 00 " + ;
"00 00 80 00 00 00 C4 0E 00 00 C4 0E 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 80 00 00 80 " + ;
"00 00 00 80 80 00 80 00 00 00 80 00 80 00 80 80 " + ;
"00 00 C0 C0 C0 00 80 80 80 00 00 00 FF 00 00 FF " + ;
"00 00 00 FF FF 00 FF 00 00 00 FF 00 FF 00 FF FF " + ;
"00 00 FF FF FF 00 77 77 77 77 77 77 77 77 77 77 " + ;
"77 77 77 77 77 77 00 00 00 00 00 07 77 77 00 33 " + ;
"33 33 33 30 77 77 0B 03 33 33 33 33 07 77 0F B0 " + ;
"33 33 33 33 30 77 0B FB 03 33 33 33 33 07 0F BF " + ;
"B0 00 00 00 00 00 0B FB FB FB FB 07 77 77 0F BF " + ;
"BF BF BF 07 77 77 0B FB 00 00 00 07 77 77 70 00 " + ;
"77 77 77 77 00 07 77 77 77 77 77 77 70 07 77 77 " + ;
"77 77 07 77 07 07 77 77 77 77 70 00 77 77 77 77 " + ;
"77 77 77 77 77 77 "

* --------- exit.bmp ----------------
cHexexit:= ;
"42 4D 42 01 00 00 00 00 00 00 76 00 00 00 28 00 " + ;
"00 00 11 00 00 00 11 00 00 00 01 00 04 00 00 00 " + ;
"00 00 CC 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 80 00 00 80 " + ;
"00 00 00 80 80 00 80 00 00 00 80 00 80 00 80 80 " + ;
"00 00 C0 C0 C0 00 80 80 80 00 00 00 FF 00 00 FF " + ;
"00 00 00 FF FF 00 FF 00 00 00 FF 00 FF 00 FF FF " + ;
"00 00 FF FF FF 00 77 44 44 44 44 44 44 44 70 00 " + ;
"00 00 77 77 4D 50 07 FF F4 77 70 00 00 00 77 77 " + ;
"45 D5 0F FF F4 77 70 00 00 00 77 77 4D 5D 0F FF " + ;
"F4 77 70 00 00 00 77 77 45 D5 0F EF E4 77 70 00 " + ;
"00 00 77 77 4D 5D 0F FF F4 77 70 00 00 00 77 77 " + ;
"45 D5 0F EF E4 77 70 00 00 00 77 77 4D 5D 0F FF " + ;
"F4 77 70 00 00 00 77 77 45 D5 0F EF E4 77 70 00 " + ;
"00 00 77 77 4D 5D 0E FE F4 77 70 00 00 00 77 77 " + ;
"45 D5 0F EF E4 77 70 00 00 00 77 77 4D 5D 0E FE " + ;
"F4 77 70 00 00 00 77 77 44 44 44 44 44 77 70 00 " + ;
"00 00 77 77 77 77 77 77 77 77 70 00 00 00 77 77 " + ;
"77 00 00 00 77 77 70 00 00 00 77 77 77 0A AA A0 " + ;
"77 77 70 00 00 00 77 77 77 00 00 00 77 77 70 00 " + ;
"00 00 "

* --------- door.bmp ----------------
cHexdoor:= ;
"42 4D 36 05 00 00 00 00 00 00 36 04 00 00 28 00 " + ;
"00 00 0D 00 00 00 10 00 00 00 01 00 08 00 00 00 " + ;
"00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 01 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 80 00 00 80 " + ;
"00 00 00 80 80 00 80 00 00 00 80 00 80 00 80 80 " + ;
"00 00 7F 7F 7F 00 BF BF BF 00 00 00 FF 00 00 FF " + ;
"00 00 00 FF FF 00 FF 00 00 00 FF 00 FF 00 FF FF " + ;
"00 00 FF FF FF 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 00 00 00 00 03 00 00 00 00 00 00 00 00 00 " + ;
"00 00 03 00 00 00 03 00 01 0B 0B 0B 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 0B 0B 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 0B 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 0B 00 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 01 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 0E 01 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 01 0E 0E 00 0B 0B 0B 0B " + ;
"0B 00 03 00 00 00 03 00 00 00 00 00 00 00 00 00 " + ;
"00 00 03 00 00 00 "

 * --------- next.bmp ----------------
 cHexnext:= ;
"42 4D 36 01 00 00 00 00 00 00 76 00 00 00 28 00 " + ;
"00 00 14 00 00 00 10 00 00 00 01 00 04 00 00 00 " + ;
"00 00 C0 00 00 00 00 00 00 00 00 00 00 00 00 00 " + ;
"00 00 10 00 00 00 00 00 00 00 00 00 80 00 00 80 " + ;
"00 00 00 80 80 00 80 00 00 00 80 00 80 00 80 80 " + ;
"00 00 C0 C0 C0 00 80 80 80 00 00 00 FF 00 00 FF " + ;
"00 00 00 FF FF 00 FF 00 00 00 FF 00 FF 00 FF FF " + ;
"00 00 FF FF FF 00 77 77 77 77 77 77 77 77 77 77 " + ;
"00 00 77 77 77 77 77 77 77 77 77 77 00 00 77 70 " + ;
"77 77 70 77 77 77 77 77 00 00 77 70 07 77 70 07 " + ;
"77 77 77 77 00 00 77 70 00 77 70 00 77 77 77 77 " + ;
"00 00 77 70 00 07 70 00 07 77 77 77 00 00 77 70 " + ;
"00 00 70 00 00 07 77 77 00 00 77 70 00 00 00 00 " + ;
"00 00 77 77 00 00 77 70 00 00 70 00 00 07 77 77 " + ;
"00 00 77 70 00 07 70 00 07 77 77 77 00 00 77 70 " + ;
"00 77 70 00 77 77 77 77 00 00 77 70 07 77 70 07 " + ;
"77 77 77 77 00 00 77 70 77 77 70 77 77 77 77 77 " + ;
"00 00 77 77 77 77 77 77 77 77 77 77 00 00 77 77 " + ;
"77 77 77 77 77 77 77 77 00 00 77 77 77 77 77 77 " + ;
"77 77 77 77 00 00 "


* --------- astro.bmp ----------------
cHexAstro := ;
"42 4D 1E 72 00 00 00 00 00 00 36 00 00 00 28 00 " + ;
"00 00 6B 00 00 00 5A 00 00 00 01 00 18 00 00 00 " + ;
"00 00 E8 71 00 00 C4 0E 00 00 C4 0E 00 00 00 00 " + ;
"00 00 00 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FC FE FE FC FE FE FD FF FF FD " + ;
"FF FF FC FE FE FC FE FE FD FF FF FD FF FF F9 FE " + ;
"FD F5 FA F9 FA FC FC FA FC FC FA FD FB F9 FC FA " + ;
"F8 F9 F7 FB FC FA F6 FD F8 F8 FF FA F8 FE F9 F3 " + ;
"F9 F4 F2 F8 F3 F7 FD F8 FB FF FA FA FE F9 FE FD " + ;
"F9 FD FC F8 FC FD FB FD FE FC F9 FC FA F7 FA F8 " + ;
"F6 FB FA F8 FD FC FB FD FD F6 F8 F8 FD FF FF F9 " + ;
"FB FB FA FC FC FA FC FC FD FF FF FD FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FC FE FE FC FE FE " + ;
"FC FE FE FB FD FD F8 FA FA F7 F9 F9 F7 F9 F9 F7 " + ;
"F9 F9 F9 FE FD F5 FA F9 F9 FB FB F8 FA FA F6 F9 " + ;
"F7 F6 F7 F5 F2 F3 F1 F4 F5 F3 F1 EE EA F5 F2 EE " + ;
"F8 F3 F0 F6 F1 EE F5 F0 ED F9 F2 EF F8 F1 EE F5 " + ;
"EE EB F6 F5 F1 F1 F0 EC EF ED EC F0 F1 EF F3 F6 " + ;
"F4 F5 F7 F7 F4 F9 F8 F5 FA F9 F6 F8 F8 F2 F4 F4 " + ;
"FA FC FC F6 F8 F8 F7 F9 F9 F6 F8 F8 F7 F9 F9 F6 " + ;
"F8 F8 FB FB FB FC FC FC FD FD FD FE FE FE FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FB FF " + ;
"FE F9 FE FC FB FF FE FB FF FE FB FE FC FC FF FD " + ;
"FD FF FE FD FF FE FC FE FE FD FF FE FB FE FC FD " + ;
"FF FE FB FE FC FD FF FE FF FF FE FF FF FE FF FF " + ;
"FC F7 F5 F4 FC FB FD F6 F7 FB F9 F8 FC FA F5 F6 " + ;
"FD F6 F3 FF F6 EF F0 F2 E6 F1 F3 E7 EA EE E2 E4 " + ;
"E8 DC E6 E6 DA EC E6 DB F3 E7 DD F3 E4 DB EC E3 " + ;
"DF EB E2 DE EB E3 DC EE E6 DF EC E5 DC E8 E1 D8 " + ;
"EA E4 D9 EB E5 DA EF E3 DD EF E5 DE F0 E6 DF EE " + ;
"E6 DF ED E7 E0 EA E7 DF E8 E6 DE E8 E6 DE F6 EA " + ;
"E6 FA F1 ED F7 F3 EE F3 F3 ED F6 F8 F2 FA FA F4 " + ;
"FA F7 F2 FD F9 F4 F8 F9 F7 F8 F9 F7 F9 F9 F9 F9 " + ;
"FB FB F8 FD FC F8 FC FD F9 FE FF FA FF FF FE FD " + ;
"F9 FF FF FE FF FF FE FF FF FE FE FF FD FD FF FF " + ;
"FD FF FF FB FD FD FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FB FF FF FB FF " + ;
"FE F8 FF FF FC FF F8 F3 FC F8 F4 FA F5 F1 F7 FA " + ;
"F6 FC F8 F8 EC F2 F3 E9 EC EF E6 E7 E9 E3 E7 E7 " + ;
"E1 E5 E4 DA E6 E3 D5 E1 DA CB E7 DA D2 E1 D4 CC " + ;
"E4 D7 CF E2 D2 CB DF CD C6 E0 CD C6 DE C6 C0 E3 " + ;
"CB C5 D8 C4 BF D7 C3 BE D8 C4 BF D9 C3 BE DA C4 " + ;
"BF DC C3 BF D9 C0 BC D9 C0 BC DD CA C3 DF CC C5 " + ;
"E1 CF C8 DF CF C8 DE CE C7 DF D2 CA E2 D7 CF E7 " + ;
"DC D4 E9 D8 D5 E7 DB D7 E5 DC D8 E4 E0 DB EB E7 " + ;
"E2 EE EA E5 EF E9 E4 F0 E7 E3 EF EC E8 F3 F0 EC " + ;
"F4 F2 F1 F6 F4 F3 F6 F7 F5 F6 F9 F7 F8 FA FA F9 " + ;
"FB FB FA FB F9 FC FD FB FE FF FD FE FF FD FC FE " + ;
"FE FD FF FF FD FF FF FC FE FF FF FF FF FF FF FF " + ;
"FF FF FF FE FE FE FE FE FE FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FD FF FF FE FF FF FD FF FF " + ;
"FD FF FF FF FF FF FF FF FF FF FF FF FF FF FB FF " + ;
"FF F9 FD FE F6 F8 F8 F8 FB F9 F8 F9 F5 FA FA F4 " + ;
"F3 F1 E9 ED EC E2 F1 E5 D9 EE E4 DA E9 DE D6 E5 " + ;
"DA D2 E0 D3 CB DE CF C6 DB CB BE DD CD BD D4 C4 " + ;
"BD D4 C2 BB CE BB B4 D4 BE B8 D3 BD B7 CD B7 B1 " + ;
"CF BC B5 C5 B3 AC C8 B9 B0 C8 B9 B0 CD BC B3 C6 " + ;
"B4 AD C7 B4 AD CD B9 B4 CE BA B5 D7 C1 BC D8 BC " + ;
"B1 DB BF B4 DD C3 B7 DE C3 B9 DC C3 B9 DB C5 BA " + ;
"DD C6 BE DD C8 C0 E1 CB C6 DF CB C6 E0 D0 CA E4 " + ;
"D6 D0 E6 DA D4 EA DE D8 F1 E3 DD F3 E5 DF EC E3 " + ;
"DF F2 E9 E5 F3 ED E8 F1 ED E8 F2 ED EA F6 F3 EF " + ;
"F9 F6 F2 F8 F7 F3 F5 F6 F4 F7 F8 F6 F9 F9 F9 FA " + ;
"FA FA FA FC FC FC FE FF FB FD FE FB FD FE FF FF " + ;
"FF FF FF FF FF FF FF FE FE FE FE FE FE FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FE FE FE FE FE FE FE FE FE FF FD FF FF FD FF " + ;
"FE FD FF FC FB FD FC FE FE FD FF FF FB FE FC F9 " + ;
"FE FC EB F9 F3 EE FB F3 EF F8 EE EE F2 E7 EB EB " + ;
"DD ED EA DB E9 E2 D1 DE D6 C5 F0 D3 CE E5 C8 C3 " + ;
"DD C0 BB DA BD B8 D5 B7 B2 D4 B4 AE CD AB A5 D6 " + ;
"B5 AC D3 B5 AA C6 A6 9B D1 AE A4 C9 A5 9B CD A9 " + ;
"9F CB A8 9E C6 A8 9D C5 AB 9F CF AE 9E CB AA 9B " + ;
"CD AE 9F C9 A9 9C CA AD 9F CD AF A4 C3 A6 9D C7 " + ;
"AA A1 CF B3 A2 CF B2 A3 D1 B4 A5 D3 B8 AA D7 BC " + ;
"AE D8 BE B2 D7 BD B1 D5 BB AF DA BF BB D9 BE BA " + ;
"DE C5 C1 E1 CB C6 DB C7 C2 DE CA C5 E8 D4 CF EE " + ;
"DA D5 E8 DB D3 EE E1 D9 EF E3 DD EC E2 DB ED E3 " + ;
"DC F1 E8 E4 F7 EE EA F6 F0 EB F6 F4 F3 F8 F6 F5 " + ;
"F7 F8 F6 F9 F9 F9 FD FD FD FE FE FE FB FD FE FC " + ;
"FE FF FE FE FE FE FE FE FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FE FE FE FE FE FE FE FE FE FE FE FE FA FD " + ;
"FF FB FE FF F8 FC FD F6 FB FA F6 FB FA F6 FB F9 " + ;
"F1 F7 F2 EC F2 ED EB E8 EA E8 E3 E4 E9 DF DF E2 " + ;
"D3 D1 DB C4 C2 D9 BC B8 D9 B8 B5 D7 B2 AE C6 B2 " + ;
"A7 BE AB 9E B7 A1 95 BB A4 95 C0 A6 98 BB A1 93 " + ;
"B4 9A 8C B0 96 88 A0 95 81 A5 97 84 AC 9E 8B A9 " + ;
"9A 87 B0 A1 8E A3 95 82 AA 9F 8B A0 97 83 C3 9C " + ;
"86 C0 9B 85 C5 A3 8C C3 A4 8F C4 A7 92 C3 AB 95 " + ;
"C2 AA 96 D0 B8 A4 C9 B2 9C C7 AF 9B C6 AE 9A C5 " + ;
"AF 9D C9 B2 A2 CD B6 A7 CE B7 A8 CE B6 AA D2 B5 " + ;
"B0 D2 B5 B0 D9 BC B7 DD C0 BB D5 BB B5 D2 BA B4 " + ;
"D8 C2 BC DB C5 BF DB CA C1 E3 D1 CA E6 D6 CF E5 " + ;
"D5 CE E5 D7 D1 ED DF D9 F0 E4 DE F1 E5 DF EA E7 " + ;
"E3 EE EB E7 F0 ED E9 F3 EF EE F6 F4 F3 F8 F6 F5 " + ;
"F7 F5 F5 FB F9 F9 FB FB FB FC FC FC FE FE FE FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FE FE FE FE FE FE FE FE FE FD FD FD FD " + ;
"FD FD F8 FD FE F8 FC FD F9 FE FD FA FD FB F8 F9 " + ;
"F5 F8 F8 F2 F2 F0 E8 E7 E5 DD FC E2 DC E5 C9 C2 " + ;
"DB BE B7 DD BB B5 DF B9 B4 D7 AD A6 D3 A5 9E C9 " + ;
"9B 94 B4 96 85 B7 96 86 B1 90 7C B1 8C 76 AF 8A " + ;
"74 A8 83 6D B4 8F 7B A9 85 73 BE 8A 7D B9 88 7A " + ;
"B3 84 76 B5 86 78 BF 90 82 BF 8E 80 C4 90 83 CC " + ;
"97 8A B4 9D 87 B6 9F 89 B3 9F 86 B4 A0 87 B4 A3 " + ;
"89 B0 A2 86 B4 A6 8A C0 B2 96 C2 B1 97 C4 B2 9B " + ;
"C6 B4 9D C6 B3 9E C5 B1 A0 C5 B0 A1 C8 B3 A4 CA " + ;
"B4 A8 CF B2 AD CF B2 AD D5 B7 B2 DA BC B7 DA BD " + ;
"B8 D9 BF B9 D7 BF B9 D4 BE B8 D7 C1 BB DC C9 C2 " + ;
"E2 CF C8 E3 D1 CA E5 D2 CD E7 D7 D1 EA DA D4 E8 " + ;
"DA D4 E8 E2 DB EE E8 E1 F0 EA E5 F4 EE E9 F9 F5 " + ;
"F0 FA F5 F2 FB F6 F3 FF FC F9 F8 F8 F8 FA FA FA " + ;
"FC FC FC FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FE FE FE FE FE FE FD FD FD " + ;
"FD FD FD FD FD FD FD FF FF F9 F9 F9 FD F9 F8 FC " + ;
"F5 F2 F4 EA E3 F2 E5 DD ED DC D3 DB C8 C0 E5 C4 " + ;
"B0 D7 B6 A2 CE B1 9C BD A2 8E AA 92 7E 99 83 71 " + ;
"9D 89 77 9C 8A 79 9C A2 97 9E A0 94 BB BB AB BE " + ;
"BC AA C3 BF AC B9 B5 A3 BF BF AF BE BD AF BE B2 " + ;
"A6 D3 CB BE C5 C4 B6 A5 A8 99 93 93 85 BD B5 A8 " + ;
"B1 A1 95 C3 AD A2 A2 91 84 AA 99 8C A6 91 82 AE " + ;
"95 85 BC 9E 8B BF 9E 8A C4 A2 8B C3 9E 88 B9 A8 " + ;
"8E BE AD 93 C6 B1 9B C8 B3 9E C7 B0 A0 C7 B0 A1 " + ;
"C9 B1 A5 CC B4 A8 C9 AF A8 CB B1 AA D0 B3 AC D3 " + ;
"B6 AF D9 BD B6 DB C1 BA D7 C2 BA D4 C3 BA D7 C1 " + ;
"BB DB C5 BF DF C9 C4 DF CB C6 E0 CD C8 E3 D0 CD " + ;
"E3 D2 CF E4 D3 D0 E3 DA D0 E9 E0 D6 EC E3 DA F0 " + ;
"E7 DE F4 ED E4 F3 EB E4 F4 EC E5 FD F5 EE F5 F5 " + ;
"F5 F6 F6 F6 F9 F9 F9 FB FB FB FD FD FD FE FE FE " + ;
"FE FE FE FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FE FE FE FE FE " + ;
"FE FD FD FD FD FD FD FD FD FD FD F9 F8 F5 ED ED " + ;
"F4 E8 E6 EB DC D9 DC C9 C2 D9 BF B8 D1 B4 AB BD " + ;
"9E 95 BC 9D 8E A0 86 76 92 7D 6E 95 88 7A B8 B5 " + ;
"A7 D9 DF D4 E8 F4 E8 D7 E7 DC D6 D7 E1 A6 A3 AC " + ;
"BB B5 BA A4 9C 9D A1 96 98 8E 85 88 82 7C 81 8D " + ;
"88 91 93 88 84 9C 99 94 6A 71 6A F8 FF FB C5 CD " + ;
"C6 4F 4C 47 79 68 65 D8 BC BB CE D4 CF E2 E4 DE " + ;
"D0 CD C5 BD B1 A7 AE 98 8C AB 8C 7D C0 9C 8C CB " + ;
"A3 91 C8 B2 99 C8 B2 99 C8 B1 9B C8 B0 9E CA B0 " + ;
"A0 CB B0 A2 CC B0 A5 CD B0 A7 C8 B1 A9 CF B5 AE " + ;
"CF B3 AC CE B1 AA D3 B7 B0 D4 BD B5 D3 C0 B8 D5 " + ;
"C6 BD D7 BE BA D6 C0 BB D7 C1 BC D8 C4 BF DA C7 " + ;
"C4 DE CB C8 E0 CF CC E3 D2 CF E0 D4 C8 E7 DB CF " + ;
"E7 DD D3 EA E0 D6 EF E5 DB EB E2 D9 EC E3 DA F5 " + ;
"EC E3 F3 F3 F3 F4 F4 F4 F6 F6 F6 F8 F8 F8 FA FA " + ;
"FA FC FC FC FE FE FE FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FE FE " + ;
"FE FE FE FE FF FF FF FF FF FF FD FF FF FD FF FF " + ;
"FC FE FE FD FF FF F8 FF FB FA FF FB FD FD FD FC " + ;
"FE FF F9 FE FF F8 FD FC FF FA F7 FF F6 F2 EF F1 " + ;
"F1 E6 DF DC E8 D6 CF E5 CB C4 D0 B7 AD C5 AB 9D " + ;
"AA 92 7C A5 8A 6F 93 84 82 BB B3 BA DE E7 F1 F0 " + ;
"F9 FF F5 EB EB D4 BB B1 5B 54 45 85 9F 8D FF FC " + ;
"FF 65 53 52 7E 51 4D 7D 5C 4D 79 60 4C 98 61 52 " + ;
"9B 67 5A 7A 59 4A 9D 62 58 73 4A 33 B7 A7 90 D9 " + ;
"DA D1 57 5C 5A 86 83 75 82 73 59 85 6B 53 7A 67 " + ;
"58 7A 72 65 B2 B7 AE AA B3 B0 E3 EB EA D4 D9 D7 " + ;
"A9 AD A7 99 9E 95 BF A5 95 C8 A1 93 CA A8 98 C3 " + ;
"B5 A2 BC B7 A2 C2 B3 A0 CE B2 A1 D0 B2 A1 C5 B5 " + ;
"A4 CA B5 A6 D4 BA AE D0 B7 AD C5 B3 A8 CF C3 B7 " + ;
"CC C0 B4 D3 C5 B9 D6 C2 BD D4 C0 BB D6 C3 BE D5 " + ;
"C2 BD DE CB C6 DE CB C6 E1 D1 CB E0 D0 CA E8 D6 " + ;
"CF EA D8 D1 EA DA D3 EC DC D5 EC DF D7 EF E2 DA " + ;
"F0 E5 DD F2 E7 DF EA E5 E2 FA F5 F2 FB F7 F6 F7 " + ;
"F5 F4 F3 F4 F2 F7 F9 F9 F5 FA F9 FB FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FE FE FE FE FE FE FF FF FF FF FF FF FD FF " + ;
"FF FD FF FF FD FF FF FD FF FF FF FF FC FF FF FE " + ;
"FB FD FD FF FE FF FE F9 FA FF FA F9 F8 F6 EE EB " + ;
"EF E4 FA DA DB E2 C6 C6 D7 BD B7 CF B0 A1 B2 8D " + ;
"79 A8 8A 79 AE A4 9D D4 DD E1 D3 DD CC D0 D8 CD " + ;
"A0 A4 9F 7E 72 6E 80 52 4B 8A 49 40 74 4A 3E FF " + ;
"FF F5 4F 37 4F 91 93 9B 70 60 5A 9F 90 7D 95 90 " + ;
"71 98 80 62 9A 7C 63 A9 87 70 95 79 6E 7F 6E 54 " + ;
"9E 9E 86 9A A7 9F 86 8E 8D B7 AE A1 79 5E 49 9B " + ;
"7A 66 A1 7B 69 93 6C 5D 82 5F 52 73 56 4D 97 86 " + ;
"7D BE B4 AA E3 DA CD EF E3 D7 A1 BA A0 D3 D0 BB " + ;
"D7 BE AA C5 A7 94 CD AD 9A D2 B1 9E CA AE 9D C7 " + ;
"B8 A5 D5 B4 AB CA B1 A7 C5 B4 A7 C9 B9 AC D1 BB " + ;
"AF D7 BC B2 D6 B9 B2 DE C2 BB D8 C4 BF D5 C1 BC " + ;
"D5 C2 BD D3 C0 BB DC C9 C4 DD CA C5 E1 D1 CB E1 " + ;
"D1 CB E5 D3 CC E7 D5 CE E8 D8 D1 EA DA D3 EB DE " + ;
"D6 EE E1 D9 EF E4 DC F1 E6 DE E8 E2 DD EB E6 E3 " + ;
"F0 EB E8 F8 F5 F1 F5 F4 F0 FA FB F9 F9 FC FA FB " + ;
"FE FC FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FE FE FE FE FE FE FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FE FE FE FF " + ;
"FF FF FD FF FF FD FF FF FD FF FF FD FF FF FF FE " + ;
"FF FF FD FD FB FA FC F7 F9 F9 FB FC FA ED EA E2 " + ;
"EB DF D5 DF CD C2 E8 BB B8 CC A0 99 B1 8F 82 83 " + ;
"79 68 B7 C1 B5 BF D0 CD F9 FB FF 3C 30 42 AB 78 " + ;
"64 8F 5B 4B 89 5C 4E 89 69 56 9C 7C 65 A4 7F 6B " + ;
"76 5F 4F C7 CA BB E0 E8 FF 82 9D AB 89 75 74 9D " + ;
"63 57 A7 72 57 9F 78 58 8A 75 55 79 63 47 94 70 " + ;
"6A 7D 5C 49 BD AB 9A D3 D4 D0 49 4B 4C A0 8E 87 " + ;
"A2 84 73 A3 89 78 7E 72 5A 95 7E 68 97 77 60 A2 " + ;
"7D 67 98 77 63 84 69 55 6C 52 42 5B 40 32 90 9A " + ;
"8A 6B 6E 5E CC D4 C3 DC EB D6 BB BC AC C3 AD A2 " + ;
"C7 A5 9F CA AD A8 D9 AC AF D4 B3 B0 C8 B3 AB C8 " + ;
"B7 AA D4 BF B0 D8 BA AF DF BD B7 E5 C2 BF D9 C5 " + ;
"C0 D6 C2 BD D5 C2 BD D4 C1 BC DB C8 C3 DD CA C5 " + ;
"E0 D0 CA E1 D1 CB E4 D2 CB E5 D3 CC E6 D6 CF E8 " + ;
"D8 D1 E8 DB D3 EB DE D6 ED E2 DA EF E4 DC EF E7 " + ;
"E0 EA E2 DB ED E7 E0 F8 F4 EF F4 F1 EC F5 F5 EF " + ;
"F5 F6 F2 F7 F8 F4 FD FD FD FE FE FE FF FF FF FF " + ;
"FF FF FF FF FF FE FE FE FE FE FE FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FE FE FE FE FE FE FD FF FF FD FF FF FD FF FF FD " + ;
"FF FF F4 F8 F9 FF FE FF FD FD FD ED F3 EE E3 EB " + ;
"E1 E7 E3 D8 EA CD C4 EA B6 AF CA A7 93 90 81 6E " + ;
"98 A2 92 E2 F7 F4 BE D1 D9 C1 CA D8 6E 71 79 61 " + ;
"63 63 3E 3D 28 B1 93 80 AA 78 64 AC 83 6A 96 75 " + ;
"5B A8 83 6D 89 62 53 45 22 15 70 8B 99 A2 D5 D7 " + ;
"72 89 81 4F 57 46 47 56 3A 56 5E 40 61 6D 4F 62 " + ;
"69 4E 52 61 59 5F 63 50 3F 4A 36 61 78 73 E5 F4 " + ;
"F6 7B 6C 69 3E 22 17 59 43 38 84 5D 4F 96 6E 5C " + ;
"9C 6F 5A AD 7C 66 A8 79 63 A1 7D 6B 51 41 30 6B " + ;
"68 59 91 BC B3 89 88 84 54 32 32 63 4A 46 D3 D1 " + ;
"C9 CC D3 CE BF C5 C4 B2 B7 B8 C1 A3 A8 D0 B0 B1 " + ;
"D7 B5 AF D5 B7 AC D0 B9 AA CD BA AD DC C6 C0 D7 " + ;
"BB BA D7 C4 BD D5 C2 BB D5 C3 BC D7 C5 BE DC CA " + ;
"C3 DC CC C5 DF CF C8 DE CE C7 E5 D3 CC E6 D4 CD " + ;
"E6 D6 CF E8 D8 D1 E7 DA D2 E9 DC D4 EA DF D7 EC " + ;
"E1 D9 F0 E7 DE EF E6 DD EE E7 DE F6 F0 E9 F1 EE " + ;
"E6 F4 F2 EA F5 F5 EF FD FD F7 F7 F9 F9 F9 FB FB " + ;
"FC FE FE FD FF FF FD FF FF FD FF FF FD FF FF FD " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FE FE FE FE FE FE FD FF FF FD FF FF " + ;
"FD FF FF FC FE FE F5 FD FD EB F1 F0 F4 F7 F5 F7 " + ;
"EF E8 F0 DA D4 EA CF C5 D1 B7 AB 9C 86 7A 8A 9A " + ;
"7B D8 E0 CF EF EA E7 9C 8A 91 5C 48 4D EA E2 E3 " + ;
"66 71 6F EA FF FF 48 4D 44 89 6C 63 AD 79 6D 73 " + ;
"49 36 73 5E 48 7D 72 5C 90 8F 7A 9F A8 93 82 7A " + ;
"73 BE C7 BD FF FF FB DE F1 E8 CF F1 E6 EB F4 EA " + ;
"F2 F9 EC ED FF EE FD FF FF F7 F1 E6 DD E2 D3 AF " + ;
"CA C7 D6 F5 F8 9C AB A7 8B 9C 8F 72 90 83 6F 75 " + ;
"6A 82 82 76 74 6A 59 69 50 40 67 44 36 6B 48 3E " + ;
"AB 93 8D CE C2 BC D9 D8 D4 C1 AC AA 8B 6B 65 5C " + ;
"40 35 AC 96 8A DD CC C3 C4 C9 C0 97 B4 AB CD D1 " + ;
"CC C2 B7 B3 BB A4 9C D1 B3 A8 D6 BA AF CB B7 AC " + ;
"D5 C3 BC CE BE B8 D6 C3 BC D6 C3 BC D4 C2 BB D7 " + ;
"C5 BE DB C9 C2 DD CD C6 DE CE C7 DE CE C7 E5 D2 " + ;
"CD E7 D4 CF E6 D6 D0 E8 D8 D2 E7 D9 D3 EA DC D6 " + ;
"EA DE D8 EC E0 DA EF E4 DC F5 EC E3 EE E7 DE EF " + ;
"EA E1 F1 EE E6 F4 F2 EA F3 F1 E9 FB FC F3 F5 F7 " + ;
"F7 F7 F9 F9 FA FC FC FC FE FE FD FF FF FD FF FF " + ;
"FD FF FF FD FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FE FE FE FF FF FF FD FF " + ;
"FF FD FF FF FC FE FE FA FC FC FB FB FB F3 FA F7 " + ;
"E7 E7 E1 F4 DA D3 F1 BA B5 D9 AA A2 85 7F 72 A9 " + ;
"CE BA F6 FF FC 7B 6B 64 7B 4E 43 96 63 53 78 57 " + ;
"44 FF FA EC 46 49 47 DB E4 E8 FF FD FE 33 2B 24 " + ;
"54 54 46 B6 C6 B4 F8 FF F5 F7 F1 E4 EC DF D1 C8 " + ;
"C1 AE B4 CA B2 BB C8 BA D9 C5 C4 D5 CE D3 DB E1 " + ;
"E8 E7 DD E3 CA D2 D1 B0 DA CF C0 D5 D7 C6 D4 C9 " + ;
"E9 F7 EB C9 E1 E1 BE D4 DA E4 EC EC C6 CD C6 BD " + ;
"CE CA B4 C5 C1 BA C3 C0 D9 DB D5 EA EC E6 B8 C0 " + ;
"B9 49 5A 51 EB FF FA 5A 70 6B B6 90 8B 88 65 58 " + ;
"8D 75 5F 91 86 68 86 72 53 91 6E 54 96 77 62 53 " + ;
"45 32 A1 AC 9C F1 FB EB E1 E1 D5 B9 AD A3 C7 B0 " + ;
"A8 DD C0 BB D6 BE B8 D3 C1 BA D8 C4 BF D6 C2 BD " + ;
"D2 BF BA D6 C3 BE D8 C5 C0 DF CC C7 DE CE C8 E1 " + ;
"D1 CB E4 D1 CC E6 D3 CE E6 D6 D0 E8 D8 D2 E9 DB " + ;
"D5 EC DE D8 ED E1 DB EF E3 DD ED E4 DB F7 EE E5 " + ;
"EC E5 DC EB E5 DE F3 F0 E8 F7 F5 ED EC EC E6 F2 " + ;
"F2 EC F7 F7 F7 F8 F8 F8 FA FA FA FD FD FD FE FE " + ;
"FE FF FF FF FE FE FE FD FD FD FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FE FE FE FF " + ;
"FF FF FD FF FF FD FF FF FA FC FC F7 F9 F9 FA F7 " + ;
"F3 F6 F0 EB F0 E4 DA E1 C6 BC D5 B2 A8 93 7A 70 " + ;
"F1 F3 E7 E2 F9 EB 7E 4C 56 8B 68 5E 95 7C 5C 9C " + ;
"7F 5A 9C 76 58 72 51 42 E7 DD D6 9A A5 A2 4D 61 " + ;
"55 F8 FF F7 F1 FB EA F9 FF F3 BF B6 B2 BE AC AB " + ;
"D3 D4 CA D1 F1 D9 D2 F1 E8 C1 EA E3 CE DA DC AF " + ;
"B7 BE A5 B5 BB BC BB BF B7 BE BB AA BB B2 A6 B7 " + ;
"C0 AF BF B8 CF E2 D9 92 AB AD A2 BE C5 AB C7 C7 " + ;
"B2 D4 CD BE EA E4 C6 E1 DD D2 E9 E5 A9 B9 B8 96 " + ;
"A3 A5 88 96 95 F1 FB FB C7 C8 C6 85 7D 7D 7A A2 " + ;
"96 4F 40 37 A7 69 5E 96 5D 47 84 68 49 97 86 65 " + ;
"92 79 5F 99 77 60 70 60 49 7B 76 61 B5 BD AC F1 " + ;
"FB EF C5 C5 BF BA AC A6 D3 B9 B2 DE BF B6 D7 C3 " + ;
"BE D6 C2 BD D1 BE B9 D6 C3 BE D7 C4 BF E0 CD C8 " + ;
"DE CE C8 E1 D1 CB E4 D1 CC E6 D3 CE E6 D6 D0 E9 " + ;
"D9 D3 E9 DB D5 EC DE D8 EE E2 DC F0 E4 DE E9 E1 " + ;
"DA F0 E8 E1 E8 E2 DB EA E6 E1 F1 EE E9 F6 F6 F0 " + ;
"EE EF EB F2 F3 EF F6 F6 F6 F6 F6 F6 F8 F8 F8 FA " + ;
"FA FA FD FD FD FF FF FF FE FE FE FC FC FC FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FE FE FE FF FF FF FD FF FF FC FE FE F9 FB FB F6 " + ;
"F8 F8 ED EF E9 E8 D5 CE F6 D5 CC A6 96 89 80 8E " + ;
"7C DD F0 DF 9B 94 8B 7C 55 53 81 69 53 9C 79 65 " + ;
"A4 77 62 A8 7F 66 73 57 39 45 3A 1F 9B 9D 8A F9 " + ;
"FF F4 DE C5 B5 D6 C5 B2 D3 D8 C3 BD D4 C6 B2 C5 " + ;
"C2 CE D6 D6 A7 B4 AC D1 F0 DB 9E C6 D2 8E D0 D5 " + ;
"70 8F 98 C8 D2 D9 B7 C6 C8 B2 C6 C1 AD C8 BE B8 " + ;
"CC BF AB BB C7 B4 C5 C1 DF EF E8 A1 AD B3 B1 B9 " + ;
"C6 B3 BE C2 B4 C2 C0 AC BD C0 B1 C3 BC AA C3 BF " + ;
"B1 D3 D2 B5 D8 DB C5 E1 E2 9B AD AC C9 D7 D3 EF " + ;
"FC F4 90 96 A1 8F 90 94 B7 BA B1 6E 73 5E 69 5D " + ;
"45 7B 58 44 A9 82 73 8F 74 66 A8 76 5A 87 65 4E " + ;
"4F 48 39 79 8D 81 C0 DB D2 EC F7 EF D8 C5 BD E8 " + ;
"BA B2 D4 C0 BB D5 C1 BC D2 BE B9 D9 C6 C1 D8 C5 " + ;
"C0 E1 CE C9 DE CE C8 E0 D0 CA E5 D2 CD E7 D4 CF " + ;
"E9 D6 D1 E9 D9 D3 E9 DB D5 EB DD D7 EC E0 DA EF " + ;
"E3 DD F0 E7 E3 F0 EA E5 EE E8 E3 F0 EC E7 EC E9 " + ;
"E5 EF EE EA EE EF EB F1 F5 F0 F4 F4 F4 F4 F4 F4 " + ;
"F5 F5 F5 F9 F9 F9 FD FD FD FF FF FF FF FF FF FE " + ;
"FE FE FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FE FF FF FF FF FE FF FB FD FE FB FB FB FF FF FE " + ;
"F1 EE E9 FF FD F5 DC DF D6 DF CB D0 D9 A6 A3 A3 " + ;
"AB 8C CF FF FA DF D8 FD 3F 10 26 A7 8A 5E A3 81 " + ;
"69 8C 75 5F A0 86 6E 55 3D 25 8C 8B 77 F4 FF F4 " + ;
"FB FC EC A7 9B 7F BD B4 AA E8 FC FF B7 C8 D1 B1 " + ;
"B2 A9 BB C5 B5 A6 B6 BC A3 AF C1 BB C8 D0 DD D6 " + ;
"CD CC EB EE A2 C6 CE A3 B0 B2 BA C5 C3 B9 C3 CA " + ;
"BC BC C8 BD C0 C5 A1 C5 CF 8F B3 B9 C5 EA E8 AB " + ;
"CD C6 9E BD B4 B4 D0 CA C3 D8 D9 C0 D4 D9 CC DC " + ;
"E8 C9 D1 DE DD D6 E3 A5 BA C2 DC FE FF 97 AD B3 " + ;
"A1 C1 C6 C0 DD E2 AE E5 DC 68 9F 96 97 BF BD CF " + ;
"DA E2 E8 F9 FF 73 70 72 63 2A 21 AC 7C 64 82 6A " + ;
"40 48 3E 1C FD FF FE E7 E0 CD 5A 48 47 77 80 84 " + ;
"E6 EB F4 DF D1 DD D1 B5 B4 D9 BD BC D9 C4 C3 E7 " + ;
"C8 C9 EB C3 C4 DE C8 C2 D8 D3 C4 E6 D7 C7 E5 D2 " + ;
"CD E6 D3 CE E9 D9 D3 E6 D6 D0 EA DC D6 F1 E3 DD " + ;
"E8 DC D6 EB DF D9 F2 E6 E0 EE E2 DC F1 E5 DF F5 " + ;
"E9 E3 F3 E8 E4 F4 EE E9 F6 F2 ED F1 F1 EB F4 F4 " + ;
"F4 F7 F7 F7 FA FA FA FA FA FA FB FB FB FE FE FE " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FE FD FE FE FE FD FC FE FD FF FF FF FF " + ;
"FF F9 F7 F6 FF FF FB EC E9 E1 FB DC DB C9 AF A9 " + ;
"A1 9A 8B E1 EA DD E7 E6 F0 A2 B6 CF 82 B2 BE 1C " + ;
"22 21 AB 72 5D 9C 5B 4D 50 31 28 97 A5 9A FA FF " + ;
"F7 C5 B9 AD A1 9E 8F ED FD EC F8 FC FF 80 99 9D " + ;
"AB C2 BE B8 CB B8 B1 CE B5 AF C8 C4 BB C5 CF A4 " + ;
"AD B6 D3 DF DF 8E AC B7 CC DF EE A9 AC B0 BD CD " + ;
"C6 AD C3 C1 B0 BF C2 AB C0 BD BE CC D8 B9 C6 CE " + ;
"E9 F6 F4 D5 DD D6 C9 CD C1 C5 C5 B9 C0 B8 B1 BB " + ;
"B0 AC A1 9A 8B C1 B8 AB B4 A2 9B A1 AE A6 D5 EE " + ;
"EA AA B5 B9 D3 E3 EA C3 CC D6 BD D1 E2 CB E7 EE " + ;
"BD E6 E1 94 CC C1 AD A1 A1 FF E8 E5 69 7E 68 3C " + ;
"32 1A 54 3F 23 8C A2 86 B6 B8 C0 76 87 72 FC F8 " + ;
"F3 46 21 25 68 71 6E CB F6 EB DA EF E6 D1 BA B8 " + ;
"E9 BD BE DB C2 BE D6 CC C5 E0 D0 C9 E0 CD C6 E2 " + ;
"CF C8 E6 D3 CE E8 D5 D0 E7 D7 D1 E5 D5 CF E8 DA " + ;
"D4 EE E0 DA EA DE D8 ED E1 DB EF E3 DD ED E1 DB " + ;
"EF E3 DD F2 E6 E0 F5 EA E6 F6 F0 EB F6 F3 EE F4 " + ;
"F4 EE F3 F3 F3 F6 F6 F6 F9 F9 F9 F9 F9 F9 FA FA " + ;
"FA FD FD FD FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FE FF FF FF FE FD FF FD " + ;
"FF FF FD FD FD F6 F4 F3 F4 F0 EB F1 EB E4 DD BF " + ;
"AC 9E 96 8F FA FF FF 94 7F 7E 6F 35 30 AC 97 95 " + ;
"BD DA E1 98 99 B5 3B 38 34 6E 72 66 FC FF F2 DF " + ;
"DF D3 B0 AD 9E DF DE CA F2 F4 EE BE C3 D2 79 92 " + ;
"AC CB E2 EA 96 9B 9C C5 C4 C6 B2 BE BE AB BA C3 " + ;
"AE B7 C1 B8 C4 BE AE CC C7 BA D5 DF C0 C5 D4 B7 " + ;
"BB BC AC C7 BD 9F BD B8 BC CE CF C6 DB D8 E3 D4 " + ;
"DC B5 A7 A9 A5 92 8D 85 6F 63 89 6F 5E 8A 69 56 " + ;
"8D 67 55 93 6A 5B 7D 68 48 7B 67 48 69 50 36 C5 " + ;
"C8 AF C5 CD BC 59 47 40 91 7E 7B 95 78 7B AA AC " + ;
"AD D0 CB CC F1 EA ED C8 C9 CD C4 DD DF BD DD D8 " + ;
"C4 D1 C1 BF C6 AB B6 AF A6 64 5B 57 9D A1 B4 B8 " + ;
"9C 95 8A 5A 58 91 75 6E A1 74 69 75 4E 3F 8D 9E " + ;
"91 F9 FF F4 D2 C2 BB DB C3 BD D0 C4 BE D8 D2 CD " + ;
"DE CA C9 EC CD CE E5 D2 CD EA D7 D2 E6 D6 D0 E7 " + ;
"D7 D1 E9 DB D5 EB DD D7 ED E1 DB ED E1 DB F2 E8 " + ;
"E1 F3 E7 E1 F0 E4 DE F1 E5 DF F6 EB E7 F4 EE E9 " + ;
"F1 EE E9 F1 F1 EB F3 F3 F3 F5 F5 F5 F7 F7 F7 F8 " + ;
"F8 F8 FA FA FA FC FC FC FE FE FE FE FE FE FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FE FF FD FE FE FE " + ;
"FD FF FF FF FE FF FD FD FD FF FD FC EF E9 E4 F0 " + ;
"E8 E1 90 97 74 EC FF FF A7 9F AA 68 46 2F B5 8B " + ;
"60 78 3D 29 E7 C1 C1 94 B1 A8 DD FF FC FF FF F7 " + ;
"CD B0 A2 BE B5 A7 FF FF F7 E8 C3 BF FA DB DE AC " + ;
"C5 C9 94 BB C9 A5 C8 CC D3 E6 F3 7D 8A A4 87 A5 " + ;
"C0 47 6D 8D 74 91 9F BB D1 B9 9A B7 A0 A5 BA B8 " + ;
"BD C2 CB BB C8 CA A4 C6 BF C3 D4 D7 E4 DC E3 B4 " + ;
"AE AF 89 67 61 80 5C 54 88 65 57 90 6B 57 AC 83 " + ;
"6D AD 81 6A AE 7D 67 A7 74 60 9C 7E 63 9C 7E 63 " + ;
"96 72 5A AD A1 85 A0 95 7A 8D 68 52 A1 7F 68 8E " + ;
"67 51 8C 56 4B 82 5D 4F 7D 7A 6B A6 A8 A2 DD FB " + ;
"FC BC DF E3 F7 EF F6 A5 B0 AE 93 A2 A4 E3 FC FF " + ;
"A3 9E BB 25 06 03 CD 98 95 9B 79 6C 7E 6D 53 A1 " + ;
"6C 51 6E 54 44 A2 AD 99 F2 FF F5 DD D7 D0 D5 C6 " + ;
"C4 DE CC CD DB C1 C1 EC D0 D0 E3 D1 CA EA D8 D1 " + ;
"E6 D6 CF EB DB D4 EB DE D6 E9 DC D4 EE E3 DB EB " + ;
"E0 D8 EF E5 DE F2 E8 E1 EF E3 DD EE E4 DD F6 ED " + ;
"E9 F7 F1 EC F3 F0 EB F5 F5 EF F3 F3 F3 F5 F5 F5 " + ;
"F7 F7 F7 F8 F8 F8 FA FA FA FC FC FC FE FE FE FE " + ;
"FE FE FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FE FC FC FC FC FE FF FF FE FF F8 F6 F6 F9 F4 F3 " + ;
"F2 E9 E5 C2 B8 B1 F9 F3 EE E3 D2 D5 59 30 27 A9 " + ;
"80 59 9A 7B 54 B6 82 76 2D 0F 04 A9 E1 B6 FF FF " + ;
"F5 A8 88 83 D7 C5 BA EB FC E1 C8 C3 AA ED C3 BC " + ;
"E3 D1 D2 94 B1 AE D7 DC DD 90 A3 A0 D0 E4 EF B3 " + ;
"C9 D5 43 66 70 1A 45 60 84 9D B1 CD D0 B7 B3 CB " + ;
"AD B9 C9 C2 B6 BA C5 B0 BB BF C8 DB D8 D0 C2 C6 " + ;
"A4 7C 7E 72 54 49 A0 7A 62 A5 7F 67 A8 83 69 9B " + ;
"76 5C 9D 78 5E 9C 78 60 A5 80 6A 9A 75 61 A6 7B " + ;
"6C A9 7E 6F A6 75 65 84 6A 52 8A 74 58 AC 84 68 " + ;
"9A 81 61 99 83 60 B0 87 7E 9E 87 71 70 52 35 B0 " + ;
"B2 94 D1 E7 D5 D7 E3 E7 6E 87 97 F0 F1 FF D0 FC " + ;
"FD 54 85 87 D3 E1 FD 9F AA A2 15 0E 05 8A 68 5B " + ;
"A7 8B 73 92 83 62 9D 78 5E 64 49 35 90 82 76 F5 " + ;
"FC F5 D2 DF DD CD BB BC F4 CE CE E8 D2 CC E4 D2 " + ;
"CB E9 D7 D0 E6 D6 CF EC DC D5 EB DE D6 E9 DC D4 " + ;
"EE E3 DB EC E1 D9 EF E5 DE F3 E9 E2 EF E5 DE EF " + ;
"E5 DE F8 EF EB F6 F2 ED F3 F0 EB F5 F7 F1 F5 F5 " + ;
"F5 F5 F5 F5 F7 F7 F7 F9 F9 F9 FB FB FB FD FD FD " + ;
"FE FE FE FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FE FD FD FD FD FF FF FF FE FF FD FB " + ;
"FB FF FF FE D9 CE CA D8 CC C6 D6 BF CD 68 2F 26 " + ;
"96 6D 4D 9D 78 5E B0 7D 73 51 3D 32 A1 AB 9A F4 " + ;
"FF EE AE 94 84 D4 E3 DB CC E6 D9 E0 D1 B7 FD D7 " + ;
"B4 DE D5 BA BB CC C8 A7 B8 C1 BB B9 B9 D6 DA D4 " + ;
"7F 7B 7A 79 64 55 81 72 52 A2 A7 A6 60 6E 80 BC " + ;
"C4 BA B5 C8 B7 AF BC BE BE BB CA ED E6 E9 AA A1 " + ;
"98 7C 57 4F 9C 68 58 A4 88 66 A5 82 61 A1 80 5F " + ;
"A1 82 63 9D 7F 64 9A 7F 65 94 7A 62 9B 83 6D 92 " + ;
"7A 66 96 6F 60 9F 7B 69 A9 7D 6C 94 7C 66 9A 7E " + ;
"66 A2 76 5F 92 77 5D 97 84 69 95 7B 63 99 7A 63 " + ;
"6A 66 4E E0 F2 E1 DD DB DA 69 6E 6F E6 FF FD A2 " + ;
"AD A5 C8 E0 D8 BA CF CD 84 A5 B9 C4 D4 C9 CE D0 " + ;
"CA 90 91 87 51 28 1F B2 89 7A A4 86 6B 9D 70 5B " + ;
"70 3C 30 78 5E 57 DC EA DF EA FB F0 CC C1 B9 F0 " + ;
"CC C6 E8 D5 D0 E8 D5 D0 E6 D6 D0 EA DA D4 E9 DB " + ;
"D5 EA DC D6 EE E2 DC EE E2 DC EF E7 E0 F5 EB E4 " + ;
"F2 E8 E1 F2 E8 E1 F7 EE EA F4 F0 EB EF EF E9 F2 " + ;
"F4 EE F7 F7 F7 F6 F6 F6 F6 F6 F6 F9 F9 F9 FC FC " + ;
"FC FD FD FD FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FE FD FF FF FD FF FF FB " + ;
"FA FC FE F9 FA FE F6 F6 9E 93 8F F5 E7 E1 7C 69 " + ;
"62 88 5E 47 94 82 5D 95 7F 66 6E 51 43 37 3E 29 " + ;
"FD FF F2 A5 8D 87 DA D6 B9 FD FF F7 A9 9E A0 EE " + ;
"C8 C6 FF D8 C9 E1 C0 AD DB D9 CF 9C C4 C3 B1 C7 " + ;
"CD BD D2 CF 98 9D A0 C5 B0 A1 7C 5E 2F BA B3 9A " + ;
"8F A8 B2 8C B3 B5 BA CC C5 AA B6 BC EC E1 EB A0 " + ;
"89 81 7B 63 4D AE 83 6E B6 85 6F 9F 87 63 9E 80 " + ;
"63 99 7B 60 99 7B 62 9B 7C 65 99 7D 65 8E 73 5E " + ;
"94 7C 66 9A 82 6C A4 7F 6B 91 72 5D 9B 78 64 92 " + ;
"82 6B 93 7C 66 A4 75 60 98 78 61 9B 81 69 9A 78 " + ;
"5A A2 75 60 81 62 59 96 82 87 BD E3 E5 CB FD F3 " + ;
"B9 9E 8A 4D 28 02 AC 92 82 E3 F6 E5 BA CA D7 BC " + ;
"D6 C8 A3 B1 AD EF FF FF 70 96 90 17 05 04 99 81 " + ;
"6D A1 82 6D AA 7B 66 99 65 54 42 2A 16 DC F0 DD " + ;
"EF FF F8 CF C4 C0 EA D7 D2 E5 D2 CD E7 D7 D1 EA " + ;
"DA D4 E9 DB D5 ED DF D9 EC E0 DA EE E2 DC EA E2 " + ;
"DB EE E6 DF F1 E7 E0 F2 E8 E1 F6 ED E9 F4 F0 EB " + ;
"F2 F2 EC F4 F6 F0 F7 F7 F7 F5 F5 F5 F6 F6 F6 F9 " + ;
"F9 F9 FC FC FC FD FD FD FE FE FE FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FC FF FD FB FD FD " + ;
"FA FC FD FD FC FE FF FC FD EB E3 E3 95 89 85 EB " + ;
"DD D7 52 16 00 97 67 4F A4 7D 57 9A 73 46 39 35 " + ;
"12 FF FF F4 DE B1 A4 D5 B9 96 E5 F6 E9 A4 A6 9A " + ;
"DA E8 DD 9E B6 AC E6 DA CE F4 C7 BC D6 C3 C0 A3 " + ;
"BE C8 C2 BE D1 B6 BF C9 BA CE E7 DD E6 F3 94 8B " + ;
"6A 43 38 1D CE D0 DB 5B 68 76 B7 C5 C1 D7 E2 E6 " + ;
"A5 97 98 70 55 40 95 7C 5A A7 82 68 A9 7E 6B 99 " + ;
"7F 67 9C 7D 68 A4 85 70 A2 82 6F 98 78 65 9C 7B " + ;
"67 99 79 62 95 76 5D 92 74 59 B3 78 69 A5 7A 67 " + ;
"97 72 5E 89 7E 63 89 7D 61 A5 7F 67 9C 84 68 8A " + ;
"75 59 AB 73 6E A0 73 66 AB 7F 6E 7B 65 53 65 4F " + ;
"43 C2 A4 99 B2 A1 8E 90 69 53 7B 5D 42 B3 A2 8D " + ;
"DF E8 F1 A7 CF BD AB E4 DB 81 B3 B3 DC E7 EF 8C " + ;
"8D 97 0E 05 00 82 64 53 AB 8B 6E A4 84 61 6E 3E " + ;
"22 3B 1B 08 9E B2 A5 CB FF FB E9 D6 D1 E2 CF CA " + ;
"EB D8 D3 EB DB D5 EA DC D6 EF E1 DB EA DE D8 ED " + ;
"E1 DB EB E3 DC EE E6 DF F2 E8 E1 F2 EA E3 F2 EC " + ;
"E7 F1 ED E8 F0 F0 EA F0 F2 EC F7 F7 F7 F5 F5 F5 " + ;
"F5 F5 F5 F9 F9 F9 FC FC FC FD FD FD FE FE FE FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 F6 FF " + ;
"FF FF FF FB F5 FD FD F7 FB FC FD FF FF CF BA A4 " + ;
"C8 F6 E0 B3 DB DA D4 D3 DD 7E 8D 85 37 25 1E 58 " + ;
"26 20 FF FF EF D5 C8 CA C9 B5 BA F2 FF E9 E4 C1 " + ;
"B7 D7 C5 B4 CB CE BF DF E9 E3 C4 BE B7 EE DD D0 " + ;
"D0 CE C6 AE CB D0 C7 C3 CE A1 C8 CA AE CC CD C3 " + ;
"C2 C4 E4 E5 E3 84 81 7C 7F 80 77 9C B9 AA 5B 75 " + ;
"99 C5 BF D0 77 56 4D 9D 78 62 A1 7C 66 A8 7E 6B " + ;
"AE 86 6A A9 87 63 A4 82 6B A2 80 68 A0 7E 66 9D " + ;
"7C 62 A1 81 64 A8 86 68 A3 81 63 A3 82 61 8E 78 " + ;
"5C 93 72 5F A4 76 6F A1 72 74 95 6F 6F 92 74 69 " + ;
"99 7D 65 9D 7E 5D 8F 7B 62 90 7C 63 96 77 60 A5 " + ;
"7C 66 8E 6E 57 7C 6B 51 99 83 6A 9D 74 5E 9A 73 " + ;
"64 8F 6D 56 7C 64 4C DF CE C5 B9 CC D1 9D B8 C6 " + ;
"D7 DA E2 B8 C4 BE E8 EB FA 4E 3F 36 6D 53 35 79 " + ;
"65 4C C0 B5 B1 AE A1 A9 2C 19 14 AC 94 7C E9 FF " + ;
"F1 EC DC D6 EF D3 D2 EA D9 D0 F1 E1 D1 EE E7 D4 " + ;
"E9 E7 DC F4 E2 E3 F2 E9 DF ED E6 DD ED E8 DF F2 " + ;
"EC E5 EF ED E5 F2 EF EA F5 F4 F0 F4 F5 F1 F5 F5 " + ;
"F5 F5 F5 F5 F6 F6 F6 F8 F8 F8 FB FB FB FE FE FE " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 F6 FF FF FF F6 F8 FF FF FE F9 FC FF F3 FF " + ;
"F3 A8 8E 82 91 9F 8D EB F9 F5 3C 41 4A A2 D8 D1 " + ;
"9F CE D6 A2 A3 AD F6 EB DD 88 81 6E F9 F8 EE B6 " + ;
"DE DC C9 B4 AC E6 D3 C4 CC C1 B3 D4 D1 C9 C8 C4 " + ;
"B9 DD D5 C4 CD CD C7 B9 C9 D5 B3 BF C3 B5 D1 D2 " + ;
"9A 9F A2 B9 AE B1 B9 C7 C3 E8 FA F3 53 4D 48 EC " + ;
"E0 DC 8B 8A A4 53 3C 41 A4 79 68 A8 82 64 92 72 " + ;
"5B 8C 6A 5D 7A 57 4D 75 58 4A 6C 5A 49 74 60 4F " + ;
"79 62 52 71 58 48 6F 53 42 73 55 44 7C 5B 4B 8A " + ;
"68 58 9C 71 5C A1 74 5F A5 78 63 93 6E 58 A1 88 " + ;
"6E 90 7B 5C 8D 74 52 9C 7C 59 A3 72 5C 99 73 5B " + ;
"9C 78 60 A5 7C 66 A1 7F 68 8E 74 5C 9B 74 5E AE " + ;
"72 62 9E 75 66 97 7C 62 8D 6E 55 63 53 46 A8 BA " + ;
"BB DD F3 FE DE EC F2 CE DF DC 84 84 8A F5 FF FA " + ;
"01 14 01 BA C0 AF FF F9 F2 D6 CA C4 F8 E2 D6 57 " + ;
"31 1F 8D 8A 7B FB FF F4 E5 D4 D1 F5 DC D8 E5 E2 " + ;
"D4 E4 E1 D2 EE DC D5 F0 E4 E2 F2 E9 E0 ED E6 DD " + ;
"ED E8 DF F2 EC E5 EF EC E7 F2 EF EA F4 F3 EF F4 " + ;
"F5 F1 F6 F6 F6 F6 F6 F6 F7 F7 F7 F9 F9 F9 FC FC " + ;
"FC FE FE FE FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FA F3 FE FF FF FD F9 F8 " + ;
"FA FF A3 AD 8F 73 6D 6E 52 3F 32 CC AF A8 DD BA " + ;
"BE B6 D8 D1 8E C7 D6 E8 FF FF 9C 83 6F E3 D1 B2 " + ;
"E3 CA B6 DC DD E1 C8 C2 BD EC D9 CC E2 C8 BA D9 " + ;
"C9 BD E6 DD CF E3 DC C9 C2 C2 BC B8 C3 D1 B0 C5 " + ;
"C3 C0 D0 CF B1 A9 AA 96 88 8A AA BD BA CC E8 E2 " + ;
"BB B1 B1 6C 43 48 DF DE E0 2A 1A 0E 78 61 47 87 " + ;
"75 58 7A 6E 5C BD B1 AB CB BE BC F2 E5 E3 E3 E2 " + ;
"D4 E3 E3 D5 E6 E6 DA E2 E4 D8 DE E0 D4 C8 CC C1 " + ;
"A3 A7 9C 8A 90 85 5C 70 59 6B 6D 57 81 6B 52 85 " + ;
"5D 44 86 5A 3D AA 80 63 A6 7F 63 97 72 58 92 77 " + ;
"5C 8E 7E 61 8A 79 5E 95 73 5B 9B 72 5C 9D 78 62 " + ;
"95 6D 5A A9 79 67 99 70 61 86 71 56 9E 79 5D 6D " + ;
"5A 45 DB EA E2 B8 C8 CE A7 BD C2 B0 C6 C4 FF F7 " + ;
"F7 94 9A 99 ED FF FF D7 E4 E2 58 4B 43 65 3D 31 " + ;
"89 5B 4A B0 85 70 68 46 36 B0 B3 A4 F5 FC F5 D8 " + ;
"CF CC E9 E3 DE E9 DE D6 FA E8 E1 EA E4 DD F1 E8 " + ;
"DF ED E6 DD ED E7 E0 F2 EC E5 F0 ED E8 F2 EF EA " + ;
"F4 F3 EF F4 F5 F1 F7 F7 F7 F7 F7 F7 F8 F8 F8 FA " + ;
"FA FA FD FD FD FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF F6 E7 F6 FF " + ;
"FF FF F9 F9 FB FF 8C AA 81 BE BB C4 62 3F 31 87 " + ;
"54 44 6F 2E 25 D1 BE AF A9 B0 AD A6 A5 90 D1 C4 " + ;
"A4 F8 EA D4 F4 C3 AD FA CA BE E9 E5 E4 CE C0 B4 " + ;
"ED D6 C6 EE D9 CA E3 CF BD E9 D6 C1 D0 CC C7 B5 " + ;
"C4 D4 BF CD C9 C5 D3 CF AB AC AA A8 A0 A1 BB C6 " + ;
"C4 AA BC BB EE ED EF 82 68 6E 2B 36 10 B0 AE 90 " + ;
"C6 B6 A9 C0 AC B1 B3 A2 B0 93 85 8B A7 9A 8A 8F " + ;
"83 5F 9A 73 65 94 6F 61 8C 69 5C 85 65 5A 8C 6F " + ;
"66 9F 84 7A A9 92 8A B6 9F 97 8D 96 89 97 9E 8F " + ;
"B6 B9 AA C6 C5 B7 AC A3 96 52 3C 31 6B 46 3E 8F " + ;
"60 5C 9C 76 5E 91 7C 61 86 78 5C 84 68 50 95 70 " + ;
"5A 92 76 5E 8F 79 60 89 71 59 94 70 60 9A 83 69 " + ;
"82 5D 41 BA A3 89 FC FF F2 AD B8 B5 E5 F7 F6 B9 " + ;
"D4 D1 B9 BE BD C4 D2 D1 92 AB AD A6 BD BF 53 51 " + ;
"50 8E 71 68 A8 89 72 86 76 52 9D 6F 5D 71 5C 4D " + ;
"EC FF F6 D4 EB E6 E5 D3 D4 F9 DF DF EB E6 DD EE " + ;
"E8 DD F1 E8 DF EE E7 DE EE E8 E1 F2 EC E5 F1 EE " + ;
"E9 F3 F0 EC F4 F3 EF F4 F5 F3 F8 F8 F8 F8 F8 F8 " + ;
"F9 F9 F9 FB FB FB FD FD FD FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 F8 FF " + ;
"F7 FF F7 FF FF FF F7 D1 CF CE 4F 95 6D CE AE B9 " + ;
"5F 3E 2A A5 79 61 AE 83 6E 30 15 00 F8 E7 D2 B7 " + ;
"A5 88 E8 FB E2 EE FB F3 B7 A7 96 E3 CE B9 ED E5 " + ;
"E5 ED E5 D8 D3 CC B9 E4 D4 C4 EE CC BC FF E0 CE " + ;
"D1 C9 C2 9D B7 C3 BF C9 C3 BC CA C6 AD B8 B5 BF " + ;
"BF BF CC CA CA B6 C2 C2 99 A9 A8 F6 F9 FD 9B 7D " + ;
"88 B2 97 93 43 2A 16 48 30 12 82 6A 4E 85 6A 56 " + ;
"92 6B 62 9A 6A 66 95 79 5B 95 78 5D 95 78 5D 8F " + ;
"71 58 8F 70 59 8D 6E 59 83 64 4F 81 61 4E 83 4F " + ;
"49 71 47 40 6D 52 48 99 90 83 CF CE C4 E9 EF EA " + ;
"9C A7 A5 40 4C 4E 77 3A 2C 9B 6D 5C 8F 6A 56 94 " + ;
"6C 59 8B 6B 54 81 70 56 84 73 59 90 75 5B 90 6F " + ;
"5C 95 79 61 90 72 55 73 54 3B 9F 98 84 AD B4 A7 " + ;
"8F 8F 89 D1 E6 DD B8 BE C3 AB C0 BD DD F8 EF 8E " + ;
"9E 9D F7 FA FF 15 11 10 62 54 3E AE 92 69 8E 67 " + ;
"51 7E 5D 4A 7A 7D 6E EA FF F6 EB E6 E5 E7 D7 D8 " + ;
"F1 EB E4 F5 EC DF F1 E8 DF EF E7 E0 EE E8 E1 F2 " + ;
"EC E7 F2 EF EB F4 F1 ED F4 F2 F1 F5 F6 F4 F8 F8 " + ;
"F8 F9 F9 F9 FA FA FA FC FC FC FE FE FE FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 F7 FF F8 FF F0 FF FB FF FC 8A 8E 7B B2 DD " + ;
"D4 73 4A 47 78 5E 40 9A 72 56 78 52 40 9E 8F 7C " + ;
"DA D6 BE A7 9E 94 E4 FF FF 89 9E 95 D6 E6 DB 85 " + ;
"93 91 AD A6 A3 EA E7 D9 C6 C6 B4 AA 9C 90 FF E2 " + ;
"D4 E9 BE AD E2 D7 CF 92 AF B6 C2 D1 CD C7 D2 CF " + ;
"B2 C0 BC 93 98 97 B7 AF B0 C2 CC CC CD E5 E3 AF " + ;
"B4 B5 A2 7E 6E 74 54 47 8A 68 62 59 32 30 83 5F " + ;
"59 A5 89 78 99 7D 5E 9F 7F 56 9A 7B 5A 9C 7D 5C " + ;
"9E 7F 5E 96 79 5A 94 77 58 90 73 54 87 69 4C 85 " + ;
"67 4A 8C 60 4F 98 6D 5A 89 5D 45 69 40 27 4D 2B " + ;
"14 54 44 33 BC C4 BA B7 D1 CB 8F A1 90 5A 62 51 " + ;
"4C 37 28 9B 6A 5C A2 6F 5F 8D 6D 56 8A 71 57 93 " + ;
"75 5A 8E 6E 5B 89 6A 53 8A 6E 56 A2 7B 65 46 33 " + ;
"1E A6 9F 8C 5A 45 36 AA AC 99 D2 D8 DD AF B9 B3 " + ;
"AD BB B0 CC DD D9 BF CC D4 E8 E8 EE 4F 37 2B 9C " + ;
"70 51 A0 86 6E 96 7A 62 67 43 31 AD 9D 90 EC FF " + ;
"F8 D4 E0 DA F3 E3 DD F1 E3 D7 F1 E7 E0 F0 E8 E1 " + ;
"EF E9 E4 F3 ED E8 F4 F1 ED F5 F2 EE F4 F2 F1 F6 " + ;
"F7 F5 F8 F8 F8 F8 F8 F8 FA FA FA FC FC FC FE FE " + ;
"FE FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FB FC FA FD FE FF FB FF FF 8B " + ;
"9E 77 D3 BB DB 37 10 00 93 78 53 94 60 49 62 3D " + ;
"23 FF FF F8 5E 5E 4E E8 E4 E3 FC FB FD C8 CA B7 " + ;
"85 9A 91 D7 E3 E7 D2 D7 CE B8 B8 A8 97 90 81 D7 " + ;
"C7 C0 F5 D7 CC F6 D6 C3 E9 E0 D3 9E B2 B3 AE C3 " + ;
"C0 CA D2 D1 CC D9 D7 8E 94 93 E3 D8 DA FA FF FF " + ;
"C2 CF CD 62 4F 52 9E 80 65 A9 93 7A 6E 56 44 93 " + ;
"73 66 38 19 0A 90 7D 68 86 75 5A 8F 7A 5B 97 71 " + ;
"59 93 6F 57 8E 6A 52 82 60 48 7E 60 45 7C 5E 43 " + ;
"75 58 3D 76 59 3E 70 4C 34 6F 54 39 6B 58 37 64 " + ;
"51 2E 68 4D 2B 68 48 2B 4A 28 11 42 23 0E 42 6A " + ;
"5F CA F8 EB C7 E1 D3 22 12 05 8B 5A 4C 99 67 53 " + ;
"91 6E 54 86 6D 4D 8A 67 53 89 71 5B 89 68 55 87 " + ;
"5F 4D 87 6C 57 6D 51 39 93 6F 57 54 3F 23 BE D2 " + ;
"C6 C2 CF C7 BB CC C8 92 B4 AE B9 DA D6 88 96 92 " + ;
"E6 E1 D8 1F 16 08 A9 9A 80 99 7C 61 B9 80 6A 6B " + ;
"46 32 B9 CB BA EC FF FB E4 DC D5 FF EB E3 F2 E8 " + ;
"E1 F1 E9 E2 F0 EA E5 F3 ED E8 F5 F2 EE F6 F2 F1 " + ;
"F4 F2 F1 F7 F8 F6 F7 F7 F7 F8 F8 F8 F9 F9 F9 FC " + ;
"FC FC FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 F9 FC FF FF FE FD " + ;
"E7 EA EE 96 B2 7D 9F 63 9F 66 2B 0B 9F 7C 51 8F " + ;
"56 41 A0 95 69 DF E8 E5 84 79 6B FF FC F4 BF 99 " + ;
"94 E8 DD C1 CB E5 D7 7C 89 7B BB CC BE A7 A7 95 " + ;
"EC DC D0 F5 E0 DE D7 C2 BA E8 D4 C2 F6 EF E0 C1 " + ;
"C9 C8 A9 BB BA CC D1 D2 90 9E 9D AE B9 B7 FF FB " + ;
"FD A5 A3 A2 C8 C6 C5 94 6B 70 A4 81 60 A8 8F 6F " + ;
"92 79 59 7F 58 3C 82 58 41 76 55 42 98 78 6B 98 " + ;
"6E 69 7D 60 52 81 64 56 84 65 56 7E 5C 4C 7B 57 " + ;
"45 7D 55 43 7C 52 3F 82 57 44 71 50 36 6A 4F 35 " + ;
"68 50 3A 60 45 30 64 45 2E 5F 43 25 5F 4F 2A 54 " + ;
"51 24 6C 1F 2D 0D 07 08 6E A1 93 B6 DD C8 1F 11 " + ;
"00 59 2E 19 8D 65 49 74 58 39 87 60 4A 83 70 5B " + ;
"89 62 54 8A 63 55 82 64 51 9B 6F 57 9D 74 54 83 " + ;
"5E 3C 78 73 5E D6 D8 D2 B4 BD C1 BA C1 C4 C2 CD " + ;
"C5 D4 E5 D7 C9 D8 D0 70 79 7D 36 2C 14 B1 86 6B " + ;
"AE 7C 5E 91 6E 54 79 6F 5E EA FF F4 E3 F4 E9 FC " + ;
"DE D9 F2 E8 E1 F2 EA E3 F2 E9 E5 F4 ED EA F7 F2 " + ;
"EF F6 F2 F1 F4 F2 F1 F7 F7 F7 F7 F7 F7 F8 F8 F8 " + ;
"F9 F9 F9 FC FC FC FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"EE FA FF F9 EB DF D9 AB D7 B0 7A 82 99 6C 68 4C " + ;
"5D 4F 39 37 1A 00 CE C1 BF A4 97 8F BF B3 83 FA " + ;
"E3 D4 E0 B3 9E DF C6 AC CF B4 B0 FF EE E6 E6 C7 " + ;
"AE DE E2 DD 91 99 8E E8 DA C8 F9 E0 DE C0 A6 9A " + ;
"D0 C4 AC DB EA E6 98 B3 C1 B3 C9 BD B3 D5 CB F5 " + ;
"EC EF 9B 92 88 C2 E5 DB C8 C4 C9 95 68 5B A3 86 " + ;
"67 9D 78 64 A4 87 72 85 6C 58 78 5E 52 36 1E 0C " + ;
"78 5B 40 87 68 51 6F 66 44 74 5A 42 6A 52 3C 61 " + ;
"46 2C 6A 41 28 68 3D 28 55 32 18 55 2A 09 41 1E " + ;
"0A 26 21 00 33 27 05 48 2D 13 44 34 17 39 2B 15 " + ;
"56 3D 2D 4E 3D 1C 62 45 26 7D 4D 35 2F 12 00 72 " + ;
"79 64 AF D6 C0 44 5A 47 25 18 08 86 4D 44 8E 5A " + ;
"32 7C 50 33 84 5D 41 95 69 4C 8D 65 48 80 66 41 " + ;
"8E 78 4F 98 76 58 73 48 33 C6 C5 B7 AE C4 C2 B8 " + ;
"C8 CE 9C AB AD B2 C0 BE BE C7 C4 FD FF FF 58 5B " + ;
"59 55 32 28 A6 6B 57 A9 7C 67 5D 4B 40 7A 77 73 " + ;
"F9 FE F5 E5 F1 DD FA E3 DB ED E9 DE EC F5 E8 EE " + ;
"F2 E7 F2 EE E9 FA F8 F7 F6 F5 F7 F7 F1 F6 F6 F6 " + ;
"F6 F8 F8 F8 FB FB FB FE FE FE FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FD FF F6 FC FF CD BF CB E8 FF FC 93 97 " + ;
"BA 5C 5E 5E F7 FA FF 8F 92 89 B6 EB E1 66 56 45 " + ;
"F4 D4 B1 D1 C8 BE E1 C8 AE E4 CC B0 E4 CA BA D8 " + ;
"BE A6 E3 BE A8 D5 C3 B8 CE DB CB A8 BB A8 F0 E6 " + ;
"DF D8 BD B3 BE A8 96 FF FF F7 74 90 9B C6 D5 D1 " + ;
"F5 FF FD B7 A9 AD 8F 89 82 A2 BE B8 CC C7 C8 9C " + ;
"63 5A AF 6E 60 98 83 68 81 7D 65 85 84 6F 89 74 " + ;
"59 7E 55 3E 38 19 02 46 20 02 5A 1A 08 44 09 00 " + ;
"2C 06 00 10 00 00 0F 09 0A 00 0E 01 00 23 0D 08 " + ;
"27 12 3A 31 28 2A 34 24 17 19 0D 0B 06 00 00 07 " + ;
"00 00 01 00 05 00 00 0E 16 00 38 21 0B 42 35 1F " + ;
"57 36 27 30 18 0C 1B 13 06 BB CE BF 89 8D 82 01 " + ;
"07 00 58 48 2B 37 38 11 28 20 0F 54 37 30 91 6D " + ;
"4F 8E 6C 4E 8A 68 51 93 71 4D 8A 65 4B 9F 9D 8B " + ;
"A3 C4 BD 85 9C 9E C0 C5 C8 BC D3 CF C3 E2 DB 9D " + ;
"9B 9B F3 FF FF A1 B2 A9 24 28 1C 41 38 34 F4 FB " + ;
"FF 99 BD C7 C7 EB EB E4 F0 EA F1 E6 DE EE EC E2 " + ;
"EE F2 E7 EF F0 E7 EF EC E7 F5 F3 F2 F0 F0 F0 FC " + ;
"F9 FB F7 F7 F7 F9 F9 F9 FC FC FC FE FE FE FF FF " + ;
"FF FF FF FF FF FF FF FE FE FE FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 F9 FE FD F6 FF FF DE D3 CF AB " + ;
"B8 B0 D1 C9 D4 C2 BF BA 73 7B 82 9D B2 A9 B0 D9 " + ;
"D2 4B 55 31 FE FE E6 C2 B4 B6 F5 C6 B1 E1 C6 B2 " + ;
"E7 D3 BA EF CA A4 E8 CE B6 EB CA B7 D8 DD C8 BA " + ;
"D6 C3 CA C4 BD FD E3 DD B0 99 89 DA B4 A2 AA B6 " + ;
"C2 C9 D9 DF B9 C1 C1 D0 CD CF D0 C5 C7 A1 A3 A3 " + ;
"DB DD D7 98 61 5E 94 75 66 9F 76 5F A2 77 64 A8 " + ;
"76 64 8C 64 48 59 3B 22 0C 05 00 3F 7C 56 4B 9E " + ;
"95 A3 EA E0 BC F8 F7 D4 FF FF E3 FC FF EF FF FF " + ;
"F1 FF F5 FD FF F8 FF F9 FF EE FD FF F7 FB FF FF " + ;
"FC FF F6 FF FF C6 D0 DA 8E 90 9B 41 48 45 03 0D " + ;
"01 02 04 00 08 04 00 2E 17 0F 3A 17 14 14 05 02 " + ;
"AA BE B9 77 90 8C 00 06 00 C1 E4 BC F0 FF FB 8A " + ;
"79 82 5C 3F 20 84 62 45 87 63 59 86 69 42 80 63 " + ;
"44 78 73 5E EA FF F9 C2 DD DA EE F8 F8 C3 DC D8 " + ;
"B3 D4 CD E6 EB E9 6F 85 83 E5 F6 F2 D6 E6 DF EC " + ;
"FF FF 38 4B 5A C9 D8 E8 35 3D 44 EA F1 EA EB F3 " + ;
"EC EF EF E9 F2 ED EA F1 EE EA F7 F8 F6 FA F8 F7 " + ;
"F8 F6 F6 F9 F9 F9 F9 F9 F9 FB FB FB FD FD FD FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FE FE FE FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FD FF F8 F8 FF F7 " + ;
"EB E8 CC 9D 96 87 4A 2C 13 A6 8C 74 B8 A4 99 99 " + ;
"9A 80 8E 9B 99 77 71 42 FD F7 F0 DD D9 E5 C8 B3 " + ;
"A4 DD BA B6 EB C4 B5 EC CE AB E8 DC C0 E8 C9 B4 " + ;
"E9 D7 C0 CE CC BA DD D1 D1 E4 D5 D2 B2 A3 90 E0 " + ;
"BE A7 D9 D6 DF AD C6 D0 B0 BC C0 C2 C7 C8 AB A0 " + ;
"AA 95 92 94 BF D0 C5 8E 5D 5F A0 7C 64 A4 7E 6C " + ;
"9D 66 51 7E 45 2F 77 5C 4E 06 19 0A A5 D4 C4 B6 " + ;
"DB D9 FF FC FF F7 EC FC E0 D1 DF B6 A8 AE 9C 8D " + ;
"91 5E 51 59 2C 24 2B AD A9 A8 54 4C 6A 53 66 69 " + ;
"66 6E 6E 5F 50 65 86 8A 9D 58 6B 70 D3 DF E1 D7 " + ;
"D9 E3 F1 FF FE 77 84 82 8D 9F 98 34 33 2F 04 03 " + ;
"00 13 00 00 04 00 03 9B AB B1 4E 5C 5A 8C AC 94 " + ;
"EE FF FB B3 A5 A6 39 20 00 6C 50 31 79 5C 4D 88 " + ;
"70 4C 8B 70 4E 54 45 2B BF C9 B9 A9 C0 BB A4 B9 " + ;
"B7 98 A8 A7 98 A6 A4 BC CD C9 D3 E0 DE A1 A3 9D " + ;
"3F 43 38 F4 FD F3 C8 C3 C2 D6 BA BA 5A 3D 34 CA " + ;
"BF AB ED FF FC E5 E6 E2 F9 EF EF F4 EF F0 F2 F4 " + ;
"F4 EF EF EF F8 F6 F6 F2 F7 F5 FC FC FC FD FD FD " + ;
"FE FE FE FF FF FF FF FF FF FF FF FF FF FF FF FE " + ;
"FE FE FF FF FF FF FF FF FF FF FF 00 00 00 FF FD " + ;
"FF FB FF FF E3 E4 D0 A1 8D 8C 60 32 13 89 5F 48 " + ;
"80 53 46 7E 6A 51 81 7B 80 A6 90 5C F1 E3 E7 B7 " + ;
"B6 C6 EC FD F0 9E B5 B7 C8 BF BB FB CA BA CC B7 " + ;
"98 F1 D7 C9 E9 C4 AE F6 D7 C2 A9 A8 AA ED F4 EF " + ;
"CD C1 A9 AD 8F 7E E8 E6 E6 B5 CF D6 BE CC D2 BF " + ;
"C6 C9 CA C8 D4 B9 CB CA B4 D7 C9 9D 72 7B 9C 73 " + ;
"53 A0 76 5F 34 2E 11 7C 9B 7E E9 FF F9 ED FF FF " + ;
"C7 D4 E4 FA DF FF 23 2E 1A 73 7F 73 59 69 5E 1A " + ;
"2E 21 AF C4 BC 6D 7B 81 B6 BE CB F6 FC FF BA C8 " + ;
"D4 D2 E9 E1 9F A7 9C FB FC FF 65 78 7F A8 BF BA " + ;
"00 07 00 8E A0 9F 00 00 04 8B 98 9A C5 CF CF CD " + ;
"E6 E2 F5 FF FF 95 99 9A 07 00 03 00 00 04 A5 AA " + ;
"AB 00 05 03 06 0B 02 42 33 20 64 4B 2B 5F 4A 24 " + ;
"59 40 20 75 51 39 74 59 34 6D 4F 32 97 91 7E B7 " + ;
"CC C3 9D AC AE B5 BB C0 B5 C1 C1 A2 B1 AD C5 C8 " + ;
"D0 C9 DF DA F6 FF F7 1F 0B 00 80 57 48 7A 59 46 " + ;
"97 80 66 70 58 3A E7 FF F3 F0 F4 EE FB F1 F1 F7 " + ;
"F2 F4 F3 F5 F6 F6 F5 F7 F7 F7 F7 F4 F9 F7 FE FE " + ;
"FE FE FE FE FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FE FE FE FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF F9 FF F9 F9 FF DE EB DD 8D 7B 82 82 53 " + ;
"37 8A 68 50 96 6E 5C 84 7C 5F 7B 67 74 A4 91 5E " + ;
"ED FF FF 93 9D A4 AB B5 A9 B4 ED EC 98 C1 BC FA " + ;
"D1 CF C4 9C 83 C6 AA A3 F7 D2 BE E8 CA B1 E1 EA " + ;
"E7 B1 C5 B9 E6 DA C2 F4 D3 D0 FF FF F8 8C 9B 9E " + ;
"A3 BC BE B4 C3 C6 BC BC C8 A5 BA BB BE E1 D7 A0 " + ;
"8D 96 8B 62 49 46 31 12 CD E3 CA DB FE F0 F0 FF " + ;
"FD 89 7C 8C 4A 32 54 35 50 6A 84 93 8B AA BE B1 " + ;
"F3 FE F4 DD DD DD E5 E3 E2 C1 B8 AE AA 99 8C 67 " + ;
"56 4D 8E A3 94 FF FE F6 99 86 79 C1 BF AD C6 D7 " + ;
"CE CA C8 C7 F2 E9 E0 EE FF EF 92 A0 9F ED F2 F3 " + ;
"00 00 03 9A A3 A7 08 0B 0F C8 D6 D5 F0 FD FB CD " + ;
"C3 C3 00 02 00 36 3D 40 04 07 00 33 27 0F 5C 46 " + ;
"2A 64 51 2B 64 4A 22 60 2F 19 7E 64 3F 7C 52 35 " + ;
"7C 76 5F CE E3 DA B0 B2 BA B2 BB C5 98 B3 B7 BB " + ;
"C3 C3 9D A6 B4 A1 AD B3 E5 E6 E2 9B 83 77 73 44 " + ;
"36 AC 7C 6A 9C 7A 63 5C 47 2C 98 A6 94 FF FF F8 " + ;
"F3 F0 EC EE EC EC F7 F8 FC F8 F7 FB F5 F5 F5 F5 " + ;
"F8 F6 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FE F9 F8 FB FF D6 EC DA 66 " + ;
"53 56 8A 54 3D 87 69 4C 81 5C 46 7A 7F 60 74 78 " + ;
"8A C1 9E 73 F6 FF FE E4 F8 EC E5 E7 D3 D5 E5 DE " + ;
"BD D7 D1 AB AB A5 D1 B9 A1 E8 CB C4 F4 D5 C6 DA " + ;
"C2 AC E1 D7 CD E2 DF D1 BD B6 A7 F5 E4 E8 A6 94 " + ;
"83 ED E4 E1 8B AA A9 B6 CC D1 BE BD C6 B2 BF C1 " + ;
"B4 C6 C5 C2 C9 D2 44 18 0B 7C 95 7B F7 FF F6 DD " + ;
"C7 C9 B4 9F 9E 5D 54 57 AA B6 BC 81 9C 99 FB F2 " + ;
"FF D9 D7 DD A6 96 9D 9C 81 8A 92 80 81 A4 93 86 " + ;
"CE BA A8 C2 B8 A6 88 90 79 DA C8 C1 4B 25 19 FF " + ;
"EC D6 B4 B3 A9 61 50 54 AE 95 93 BF BB B0 E2 E4 " + ;
"DE B6 B7 B5 D2 CC D1 B1 9D A9 24 37 3A A6 B9 B6 " + ;
"00 07 00 BC BC AE FF FF FE 60 6D 65 25 31 31 05 " + ;
"00 00 41 34 1E 45 38 18 67 52 33 6F 4C 2A 49 3A " + ;
"13 65 3C 1C 64 5F 46 B7 D5 CA A4 AD B6 82 8F 9F " + ;
"64 86 8C B1 C3 C4 B4 C2 C1 C8 C8 C8 9E 9B 97 EA " + ;
"E8 DD 68 53 44 9F 6B 5E A3 73 61 90 79 63 59 57 " + ;
"3F E3 E7 D4 FC FF F7 F3 F5 F5 FA F9 FD F5 F6 FA " + ;
"F7 F9 FA FD FF FF FF FF FF FF FF FF FF FF FF FE " + ;
"FE FE FE FE FE FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF F8 FB FD FF " + ;
"E8 FC F0 75 55 60 89 3F 39 8E 63 52 83 53 47 8D " + ;
"8B 79 51 66 7B C6 AB 86 EB D8 D3 C7 A3 93 DB C6 " + ;
"AB C8 DD C7 E0 F1 E3 F4 DD D5 AE B8 A1 E9 CE C4 " + ;
"E5 C9 BE E5 CB BA E3 B4 A6 ED CB BE D9 DA D1 E8 " + ;
"F0 F7 DA CE B6 F5 D5 CF CF DD DC 85 95 9B CD D3 " + ;
"D8 A8 BA BB C0 C3 C8 84 8E 98 9A C4 B2 FF FE FE " + ;
"A3 91 90 AC A1 9D 52 40 41 FF FF F6 C3 B8 A2 78 " + ;
"64 52 F3 F0 EC AA B0 AF 76 79 77 99 9A 91 E5 EE " + ;
"EB 7B 77 82 72 61 76 67 5E 72 FB FF EC 8A 8A 7E " + ;
"C8 B4 A3 2F 12 03 9F 97 98 FC FD FF 95 8C 88 37 " + ;
"2D 2D EA D4 C9 BF 9B 9B 32 10 1A F4 FA FF EA FF " + ;
"FF D3 F8 F0 00 18 06 90 A3 88 7B 6F 6F 77 8A 75 " + ;
"E9 FE FF 7A 7D 92 05 06 00 1E 1A 01 41 35 23 49 " + ;
"3F 0F 49 44 1D 53 2D 0D 3E 32 1A 89 AD A1 4E 6B " + ;
"72 77 85 97 77 90 9A 82 A8 A8 AE BA A4 AA C6 B2 " + ;
"C8 EA D9 B3 C2 B4 7A 77 68 64 59 43 95 7F 66 A4 " + ;
"82 6A 67 58 3E 92 95 7F F5 FF F4 F8 FB F9 F6 F2 " + ;
"F7 FC FD FF F4 F7 FB FD FC FE FF FF FF FF FF FF " + ;
"FE FE FE FE FE FE FE FE FE FE FE FE FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FB FF " + ;
"FF FF FA FF FA FF FF 68 49 6A 7C 4B 31 A7 63 50 " + ;
"9C 62 4C 94 8D 5C 63 6A 7D BF 9B 7D F1 F7 FF E9 " + ;
"BA 9E FF C1 A5 EC B0 8C DC A5 80 E2 D1 BE C4 DF " + ;
"CF CD B1 B1 FF DE CE DC C9 A6 E5 D9 BD D6 BB A6 " + ;
"EF D6 CC D3 EA F2 C7 AC B0 7E 5F 50 EA EB E7 8B " + ;
"A4 AE C0 C4 BF C2 BF B7 83 9F 9F 98 CE C7 FF FB " + ;
"FF BE A8 C0 68 65 5D B3 BA B7 FF FD FF 58 45 3D " + ;
"A5 8F 91 FF FE FF AB C3 BB 5E 60 68 D0 BA C6 9C " + ;
"8F 97 C3 D3 D9 BC CD C9 E6 EC DB EB F2 E3 C3 C6 " + ;
"BD B9 C8 B3 EE F8 EC 66 60 5B FF FE ED E3 DA D0 " + ;
"72 5E 5D 8F 71 66 5C 57 54 5F 67 5C FF FF FF DB " + ;
"AA AC 71 2B 1A 8F 6C 5F FF F9 FF FF EB FF 00 02 " + ;
"00 07 00 07 CA BE D0 FF FB FF 84 7D 84 00 04 05 " + ;
"00 04 00 47 33 1A 4E 25 1C 51 3B 1F 2A 20 16 80 " + ;
"A5 AD 9D C3 C7 B1 BF C5 AD A3 B0 88 9F A1 A1 B5 " + ;
"B0 7F 97 95 C5 ED E1 A3 CB BF BA A8 AF 58 1E 1F " + ;
"A4 83 69 92 77 55 9D 73 49 53 43 26 E3 FF EC F2 " + ;
"FF FB FC FB F7 F7 F9 F3 FF FE FA FA FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 F9 FF F7 FF FF FE F1 FD F1 89 7D 91 7B 66 " + ;
"4B 7D 5C 49 66 46 39 69 72 51 55 65 75 B3 95 72 " + ;
"E9 F1 EA E1 C0 9F EA C3 A7 EE C8 A8 F9 CC A6 CB " + ;
"B7 94 F9 D4 C0 BA CF C0 AC BA A8 F1 CD BD EB C8 " + ;
"BE DD C5 B3 DB C6 B1 DB EE E5 B8 C1 C4 E7 E0 D1 " + ;
"E5 D5 BE EE F1 E8 8A A0 AB B4 C6 CD A4 AE A8 EF " + ;
"F9 F3 C8 CD D6 81 94 73 D7 E3 D7 F5 F0 F9 C1 AC " + ;
"B4 C3 B5 AF 7C 7E 68 C1 B5 C1 FF EC F1 FD FF F9 " + ;
"CF D2 D0 BD B8 B7 B5 A8 A0 9E 7B 77 7A 54 54 B3 " + ;
"A2 99 BE B0 B4 44 41 39 CB C3 C4 DA C5 CD 67 56 " + ;
"53 8C 87 86 FF F9 FF 83 7C 79 95 97 91 FF FF F7 " + ;
"BB AE AC 65 48 43 91 77 59 94 95 7B D8 D6 D5 CB " + ;
"A7 B3 EE DE CD F7 FD F2 53 58 57 00 04 0B ED FF " + ;
"FF FF FC FF 10 00 03 00 06 00 59 38 35 43 34 13 " + ;
"33 31 13 98 AC 9F 7D 86 7C AE BD B5 9D A7 A7 A7 " + ;
"BE B9 98 BC B6 B8 CB C8 9A BC B5 C4 E6 DF E6 E1 " + ;
"E3 66 48 47 82 67 52 9D 86 66 AC 79 59 78 53 3F " + ;
"C6 D0 C4 F3 FF FD FF FF FE FF FF FB FF FF FE FA " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FA FF F7 FF FF FC DD EA DC D3 " + ;
"D7 E2 BB C0 B1 FF FF FB AF AA B3 E3 FD F7 53 4E " + ;
"57 96 7A 57 E1 FA E6 C0 B6 98 E9 CB B2 F6 CC AF " + ;
"EB C1 9C E6 D8 AE E0 B4 8F E7 D0 C0 D8 E2 D2 B6 " + ;
"C5 B7 D1 BC BB EF CB C1 D5 C9 AD D6 CF BC BE DC " + ;
"CF BF CE C0 F4 DA BC DB BD AC A8 B2 C3 6A 88 8D " + ;
"E2 EF E1 B6 B9 BD AB 99 AA EA EC D8 CF C8 CD 68 " + ;
"4B 44 6C 4A 2C A9 96 91 FF FF FE A2 9F 91 30 00 " + ;
"07 6A 58 51 BF BC B4 D2 CF C7 49 49 39 D6 D2 C7 " + ;
"FF FF F7 E6 F2 E0 FF F5 F7 F6 E7 DE F0 DC DB F2 " + ;
"D9 DD FC F0 E6 E7 EB E5 BB C9 C8 C8 DC CF B4 C9 " + ;
"BA 3C 24 18 76 3B 3F BF 84 82 98 74 56 94 92 70 " + ;
"25 26 16 E1 CB C0 6D 32 22 98 66 54 FF F2 E7 B4 " + ;
"B5 B9 7A 78 84 61 59 60 F9 FD FF 0B 1E 26 00 00 " + ;
"00 2A 26 03 5C 5A 31 AA A1 86 5B 38 24 8E 90 7D " + ;
"A5 CD C1 91 BA AB 8F BA B7 95 A3 9F 9F BA B7 C7 " + ;
"E3 E3 EE F6 F5 BB B8 B3 3D 2C 1F 88 77 5C A3 79 " + ;
"62 84 57 4A 90 90 8A EA FF FD F8 FA FB FD FD FD " + ;
"F5 FF FE F0 FF FE FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FC FC FC FF FA FF " + ;
"FB FF FA A0 A5 AE 8D 8C 90 A8 A0 A7 B6 B3 CC A1 " + ;
"BB C9 5E 50 52 7D 5E 45 FA FF F5 D2 CE C3 AC 9F " + ;
"89 DD BE 9F ED C6 AA DA BA 9D EA CE 9F E4 BD A7 " + ;
"EC BD AF D9 DE C9 A0 BE B1 E0 C6 BA FF DA C8 D4 " + ;
"CF C6 DC C2 B1 CF CD BB C1 C3 AD DE D5 C8 E4 E2 " + ;
"E2 D7 E5 D9 D6 EF E1 A0 C4 CE A2 87 9B E9 DA EF " + ;
"73 6C 53 8E 6E 63 A1 7F 80 4D 44 23 91 93 7D FA " + ;
"FA FF E0 BE BF B0 A3 A5 96 9B 92 FE FF F6 FF FB " + ;
"FC 9A 8E 88 AF 9B 89 C2 A2 8F D1 A0 92 C8 9C 84 " + ;
"70 42 31 66 36 2A D2 AD 93 C3 A8 93 A9 96 87 FF " + ;
"FF E6 FF F9 E8 E8 D7 C2 9F 93 89 64 4A 43 95 5F " + ;
"48 8C 5D 47 7E 7F 6F D1 F9 E0 90 90 84 A4 96 80 " + ;
"6E 6E 5C DE E7 EA 1A 19 22 00 04 00 AD C0 B7 FF " + ;
"FA FF 11 1D 23 49 36 21 B4 A3 89 7B 61 53 B2 79 " + ;
"6A 4C 38 27 B0 D2 C8 D3 F5 EE C2 E1 E0 D5 DC D7 " + ;
"CD E2 E3 B5 D0 D4 B6 C2 BC FE FB F6 26 1B 13 97 " + ;
"8C 76 36 43 2B 7B 6F 63 63 78 6F CF FE F6 F2 FE " + ;
"FE FF FC FF FF FF FF FF FD FE FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FE " + ;
"FF FF FB FF F1 F1 EB E9 E8 E4 7D 6A 73 9C 7D 7E " + ;
"B0 9F AA F3 FF FF 79 75 7A 32 19 0F FA FF EA C6 " + ;
"CB CC BC C6 B6 E4 DB C7 BF AC 9F E3 C3 B0 E4 C4 " + ;
"9B F1 DB BF ED BD A1 E0 BD 9B E0 D9 C0 EB DA C7 " + ;
"CF C3 B7 C5 CA D3 F7 B7 B3 F9 D1 B8 C8 D4 BE AC " + ;
"CE C3 A8 B2 A2 FF FB EE BC D2 D0 8C C2 C3 F8 FC " + ;
"FD CA D3 DD DF E6 D1 A7 A5 A4 E6 E7 F5 FB F7 E4 " + ;
"FF FF F3 62 70 76 CD E4 E6 CE DA DE D7 D3 CE 59 " + ;
"40 3C 42 1E 24 1D 00 00 3C 13 04 C6 8A 7A CB 9E " + ;
"83 B6 8A 65 BD 8E 72 BC 90 78 A7 7F 5C 7D 51 32 " + ;
"7C 4D 37 6D 42 21 A5 77 65 CD A6 8A D6 CA B8 FE " + ;
"FF F1 DF D2 B8 81 6A 5A E1 D2 D0 D8 CE C7 B4 AB " + ;
"A7 AF B3 9A 9F 94 80 AA 99 9D D8 DD E0 FD FF F5 " + ;
"26 29 1A 00 02 04 C6 E2 E9 65 52 4D 05 05 05 97 " + ;
"AD B3 C0 BD B5 85 81 76 67 87 86 99 BC C0 AF B4 " + ;
"B5 B5 B9 B4 97 A6 A9 8E AB AF A9 AD A7 E1 CD C8 " + ;
"EA E1 DD 95 90 81 E4 F3 DE 95 85 78 9C A4 99 C3 " + ;
"ED E2 F1 FF FE FF FA FF FA F9 FB FF FF FE FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FE F6 FF FF FF ED EE DE D3 D8 BD 63 4B " + ;
"4D 8D 62 4D 6E 4D 3E 9E 91 77 90 9D AB 10 09 06 " + ;
"FA FF DC E5 F8 FB CD E8 DE BA C3 B9 C1 C7 C6 D5 " + ;
"C7 B4 F9 CA AE AB 8C 75 F1 E2 BB EB D1 A2 F1 BA " + ;
"9B EC C0 A9 DC DF D0 AD BE CB 90 9B 9F FA D1 B1 " + ;
"E5 BD A0 DF E1 D5 E1 E8 DB FD FE FC 84 9A A0 CA " + ;
"F9 E9 B9 B1 C2 99 A0 93 CA C2 D3 C9 D0 D3 C5 DE " + ;
"C4 46 38 3C BF A8 AD B2 C2 B1 FF F9 FF 5B 3A 37 " + ;
"91 60 58 F6 C2 BB 5A 36 2C 4D 3F 39 09 06 01 37 " + ;
"30 1D 30 1C 03 57 41 1E 8A 71 57 82 6B 55 6F 59 " + ;
"36 8E 69 4F AD 7F 6E C0 96 79 E8 C3 A9 FF C8 A8 " + ;
"D6 A0 89 AE 93 7F BA B8 9A EB F5 E4 7A 68 6F 43 " + ;
"00 0A 88 59 55 8F 66 4F 67 4D 35 9F 94 90 FE F4 " + ;
"F4 77 69 5D 8C 9D 92 24 6B 6E 91 B4 B8 FF FE FF " + ;
"01 0E 1E 06 2B 41 6A 86 87 99 9D 97 5B 75 7B 95 " + ;
"B8 C5 AC A9 A5 B2 B0 AF AF B8 BB A6 C2 C2 D3 D3 " + ;
"CD C5 AB A5 E3 DD D6 29 22 19 86 65 56 9B 65 5A " + ;
"3D 1B 0E 7D 89 7D F8 FF FE FF FE FF F4 FF FF F7 " + ;
"FF FD FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF F8 FD FF FF F3 F5 E9 C8 " + ;
"DD BC 8D 80 88 91 69 50 B2 86 6E 8C 62 3D A0 B6 " + ;
"C2 06 00 02 D1 C6 94 F8 FF FF C3 DE CE F0 FF F4 " + ;
"E7 F8 FB B0 A1 91 EA BD A8 97 85 74 E3 CE AF DC " + ;
"BA 8F DF C3 A0 F6 C4 A8 DC AD 98 EA FF FC A5 D9 " + ;
"D9 C5 B7 A4 F6 BE A5 CE A8 9C FE FB F7 A8 B6 B4 " + ;
"A4 B5 B1 F3 FF FC 4F 2F 52 F8 ED DF 81 7C 7B 7F " + ;
"7A 83 D4 CB C8 FE F2 E0 6C 68 45 FF F8 F6 D8 86 " + ;
"8B EA A2 90 7F 4F 33 C6 A4 8C C7 A3 93 44 28 27 " + ;
"02 00 00 00 07 00 29 13 01 5C 40 22 B2 8F 7B D6 " + ;
"B5 A6 C8 A7 8D D8 AB 96 D8 A6 9A B4 8E 76 87 63 " + ;
"45 87 70 42 71 61 37 96 75 54 7D 53 30 8C 7D 6A " + ;
"FD FF FF D4 C6 C7 60 40 35 9C 66 4F 84 6C 50 78 " + ;
"72 65 A4 76 75 47 0D 08 EE DC DB D5 E0 EE 00 04 " + ;
"0A 40 40 3A F8 FD FF 00 00 14 7B 84 87 C7 BE BA " + ;
"8D 93 9A 8E A0 AB AE BA AE BC B6 B7 C0 C5 C6 9D " + ;
"B4 AF DB E2 DB A0 9A 8F FF FF F7 64 57 55 95 70 " + ;
"5C A6 73 5F 8B 60 4B 8C 8A 78 F9 FF F9 FF FD FF " + ;
"F9 FB FB FF FF FC FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF F8 F8 FF " + ;
"FD FF FF 84 A2 89 B3 B5 CD 59 3C 2D B2 81 71 A1 " + ;
"5F 46 E5 E1 E7 42 34 38 6A 5F 2B DE DD DF C5 C5 " + ;
"AD DF CB B9 D7 DB E0 F1 EB E0 AC AB 91 D3 AF A7 " + ;
"FF E1 CF F1 C8 A8 F0 C9 AD DD BD 99 E1 C0 99 EE " + ;
"E7 CE C6 BA B4 D7 DE D9 D8 D6 C4 B1 A4 94 ED F4 " + ;
"EF AC BF AC 9F 9B 89 D9 C3 D6 90 92 7E E0 DF E1 " + ;
"C6 D9 C8 D0 D2 D3 A1 83 90 66 5F 4E FD FF EC 7F " + ;
"69 50 63 5D 34 A0 86 5E 8F 78 48 72 5D 37 7D 57 " + ;
"52 1F 02 05 00 03 00 00 13 06 A6 98 81 6C 53 31 " + ;
"96 74 5D 7C 5C 49 68 4C 2D 88 65 4B 90 70 5D 8E " + ;
"81 61 BE 92 75 B9 9F 70 DE C8 98 FF CC A4 FF C8 " + ;
"9F E7 D1 B5 3F 40 37 B6 A3 9C F6 EE E1 6C 68 4C " + ;
"A3 98 7C B5 B0 A1 C2 C6 BA EB E2 D4 C3 B7 B1 F1 " + ;
"FE FF 98 AC AD 80 A5 91 C8 F3 E6 ED FD FF 05 27 " + ;
"1D 8C A2 96 73 8D 8D 65 81 82 91 AF 9C 99 8B 8F " + ;
"A6 A8 A8 8A 9D 94 C2 D3 CA B7 C9 B8 E4 E1 D2 72 " + ;
"5D 60 87 6C 52 8F 6F 52 97 75 58 41 41 29 F8 FF " + ;
"F9 FE F9 FA FD FF FF FF FE FB FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FA FA " + ;
"FF FF FF FB FF FE FB BD CE C3 CB DE DB 77 55 5F " + ;
"BD 7F 6D 97 61 43 CF C4 A4 99 99 A7 00 07 00 E4 " + ;
"D2 B3 DC AD 8D F5 CD 9D DE C1 9C E7 C8 B9 F0 F1 " + ;
"F5 D3 C8 C4 A4 9D 8A C2 B5 9B DB B6 A0 F9 D0 BA " + ;
"D5 B2 98 FF D5 B8 DA AB 9D CB A2 93 EA E5 D6 F6 " + ;
"E5 DC C0 E8 CC D1 A4 8F FF F6 F3 4C 3D 58 E4 E3 " + ;
"E7 EF FF FF E2 FF F7 F8 FF FF 5B 53 42 EC EC D4 " + ;
"BD 9B C4 4D 17 06 91 7F 4A AC 89 5E AA 78 5A AB " + ;
"80 6D 5C 45 36 15 0F 02 35 2F 22 92 87 79 7E 6B " + ;
"45 8F 6E 54 99 6B 59 C2 97 7C F0 D0 A5 D9 BE 8C " + ;
"D2 B0 85 EC C3 A3 FE DF B2 D7 A4 84 E1 BA 93 DE " + ;
"BA 8A E8 BF 7A F4 D5 BC 99 83 78 79 47 31 8F 7C " + ;
"6F FF FA FF DB D6 DF D9 D1 D1 BA B2 BD BF AF BA " + ;
"BD A6 9E 74 5E 53 F9 F7 E5 7A 7D 8B 00 06 00 70 " + ;
"87 83 62 7D 98 2E 46 44 B9 BE AF 7C 70 82 A0 BB " + ;
"BF 9E B8 BE A8 C8 C3 88 9D 9F AD D4 DC C6 D1 C9 " + ;
"C6 C2 AF 8F 83 83 9B 70 5B 90 72 57 8D 6D 50 4F " + ;
"3D 26 D3 E1 D5 F1 FF FF FC F8 FF FF FA FF FD FF " + ;
"FF FD FF FF FD FF FF FD FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FE FA FF FF FF FB FF FD FA EE FB F3 84 95 " + ;
"92 97 83 88 9E 72 61 A9 87 6A 88 7C 58 FF F2 FE " + ;
"0F 00 09 F9 EA E1 DA BD AE EB C6 A4 F3 D0 A5 D8 " + ;
"B6 99 DA C1 A7 F2 D4 BB FC EC D5 F0 E9 D8 CE C0 " + ;
"B4 B9 AA 9A D7 C8 A8 DC C0 97 CC EA DD E2 AD 98 " + ;
"DD A4 85 C0 9D 89 F4 FF F8 85 87 81 6D 95 8A 32 " + ;
"23 27 3B 26 29 4F 4D 25 76 75 61 D4 D6 DE A1 A7 " + ;
"AC F4 F2 E0 1B 01 01 AB 95 71 A6 9F 8E BD B1 A7 " + ;
"B1 A3 A4 F1 EA F1 D5 D9 DE 00 01 01 7B 70 68 75 " + ;
"5B 4F 70 5A 37 BB 99 7C E3 B5 9D B6 87 6B E4 BE " + ;
"9B E5 C6 9F E1 C4 9F D0 AD 8C D5 B5 91 BC 8F 74 " + ;
"DA BA 97 DB BE 92 E4 C3 84 F7 D9 C8 9C 82 7C B0 " + ;
"7B 67 80 5C 4A A5 99 87 FF F5 F1 4A 25 21 7C 4C " + ;
"3A 77 53 49 6A 50 44 82 6B 3E 82 63 4A CA CA D6 " + ;
"13 32 3B 1F 51 47 DE FB FF 0B 0B 23 C2 C2 C2 99 " + ;
"95 94 9A B9 BA C8 FB FD 98 C7 C4 CF DF E5 D2 DC " + ;
"ED DE DF DB CE CC BA 83 7F 7E 82 71 5C 8F 83 67 " + ;
"92 79 59 5F 48 2E BD C1 AE FA FF F9 FF FF FB FD " + ;
"FE FA FD FF FF FD FF FF FD FF FF FD FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 F1 F9 FF F9 FD F8 FF FF FC F8 " + ;
"FF FC AA B7 B5 AD A2 A5 4D 32 24 91 7B 62 2A 0D " + ;
"00 E5 D1 DD 0E 16 27 5F 79 73 D4 D0 C5 E8 BD A8 " + ;
"FF E5 BF D9 D1 B3 DC BD 9C EA BD 97 EE BF 93 F3 " + ;
"CB A1 E9 CC AD D3 C5 B2 D9 D8 CA C4 BE B3 B7 D9 " + ;
"CE AD 97 7B AC 79 58 B2 9D 88 BF C3 BE BC CF CC " + ;
"50 7F 76 30 41 3D 6E 71 7F B9 C9 A7 BF BA B9 BB " + ;
"BF C0 E8 FF FF D7 F3 E0 B8 C7 BF DC F3 EB D8 D8 " + ;
"D2 BF BC B4 EF E5 DB B2 A2 96 85 70 61 7E 65 55 " + ;
"78 5E 4D 43 25 14 BC A7 88 72 53 34 C3 97 7A DB " + ;
"AB 8F B2 85 6A BC 97 7B DC BC 9F D9 BC 9D DD BE " + ;
"9F B9 8E 79 DB B7 99 E5 BF 9C E2 C1 89 E0 C4 B9 " + ;
"B2 9B 99 84 5D 47 A1 82 6B 75 62 3F 9C 97 82 FF " + ;
"F8 EC 63 45 2A B4 92 75 98 75 5B A1 76 4B 8A 54 " + ;
"4D C2 A5 9E F8 FD FF 13 28 37 A9 BB D2 AB B9 CF " + ;
"10 16 11 AC 9E A2 70 84 85 98 CC CC 80 B2 AE 8F " + ;
"AC B3 C5 D0 E4 DB E0 DF BE B7 A6 8F 8D 8C 60 4C " + ;
"3B 92 82 6B 90 75 5B 61 4A 34 A3 AB 9A F0 FD F5 " + ;
"FF FF FE FB FF FE FD FF FF FD FF FF FD FF FF FD " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FD FB FF FF FF FE " + ;
"FF FF FC FD FF FC D0 DA DA 7B 83 82 C1 C5 B9 96 " + ;
"9B 86 C8 CE D5 F3 FE FF A8 BB C8 00 07 00 FF FF " + ;
"F0 96 62 56 AF 78 63 CE A0 99 C6 B6 A5 ED C5 B2 " + ;
"FF CB B3 EA AB 8F CF 9A 7F A3 7D 5D A9 8C 65 CA " + ;
"A9 7B E2 C8 B8 A3 C5 A7 B4 B4 A4 D7 E3 DD ED FF " + ;
"FC EA FF F3 F4 F6 F6 2B 4B 58 00 00 01 2A 3A 23 " + ;
"33 32 1E 38 2D 0F 88 82 75 8B 8F 84 AC 98 97 8A " + ;
"63 4D B0 81 6B DD B3 96 90 69 43 7D 55 2B D0 AA " + ;
"80 FF E1 BD FF F2 D5 5E 59 40 3F 30 0F CB B1 93 " + ;
"85 5D 44 EB B9 A2 D6 A4 8D BD 8F 77 D6 B0 98 D0 " + ;
"AE 97 D3 B1 93 C6 9C 85 EE BC 9E F4 BA 97 E3 B7 " + ;
"82 F6 D7 CE A3 93 8D 78 68 4B 9F 7F 6C C0 86 73 " + ;
"77 52 38 B0 B2 A6 AB AA AC 69 53 3A 80 66 41 7D " + ;
"66 5E 6B 49 3C 3E 36 11 BA DE C8 29 3C 41 02 00 " + ;
"07 F6 FE FE 52 5A 4F C7 CF C8 77 74 76 6B 82 84 " + ;
"D1 ED ED 5B 8B 8F AA CF DD DC F3 EF CC C1 B3 97 " + ;
"8F 8F 85 56 40 94 6C 50 A6 77 58 6D 47 2F 98 94 " + ;
"82 FB FF FC FE FB FD F5 F9 FA FD FF FF FD FF FF " + ;
"FD FF FF FD FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 F6 FF " + ;
"FF EF F5 F4 FD FF FC FC FD FB FB FF FF 89 94 92 " + ;
"A4 AB A6 FF FF F8 66 59 67 9F 93 93 FF F9 FF 07 " + ;
"00 00 96 B0 9E 95 C8 B4 89 B6 9B B5 C6 B8 A9 B8 " + ;
"B4 7F 7A 6B D2 B1 97 D3 A2 82 DE AE 8C F6 C3 A3 " + ;
"DD A3 80 F4 B3 8D E6 A9 9B A4 C0 A9 C4 C9 C7 E1 " + ;
"FF FF D7 F9 F8 CB F9 EC F2 FD FF 96 AF C9 27 5A " + ;
"3E 45 66 58 4B 60 40 DF DB D0 9B 7E 97 22 15 07 " + ;
"3C 18 10 DE A5 78 FF C7 A4 AB 81 5C A0 87 5F E3 " + ;
"D0 AA C1 A8 86 ED CC B2 EC CC B5 FF EC D7 27 20 " + ;
"00 36 24 07 C4 A3 90 BE 90 7E FF CD B3 D8 A6 88 " + ;
"A1 70 5A 87 5A 4D 9E 7A 56 CF AC 8B F6 C5 9F F7 " + ;
"B8 92 EB BF 8A EF CA C2 89 6C 67 91 7D 5A B0 7B " + ;
"6E C8 7F 71 AA 68 55 5F 46 44 FD F9 FF 38 2F 22 " + ;
"EC E4 CD F1 FE FF D8 FF E9 F6 FF F9 91 8E 89 FF " + ;
"FF FC A3 B5 B4 D5 FA FF B1 C1 D8 7E 64 64 A1 9B " + ;
"9C 81 90 92 A0 AF AB DC FE FD AE CB D4 CE E1 D8 " + ;
"CD C0 B0 95 8D 8E 43 25 0C 93 76 57 91 60 40 69 " + ;
"3A 1F 99 88 75 FF FF F7 FF FF FC FF FF FC FD FF " + ;
"FF FD FF FF FD FF FF FD FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FE FF FF FF FC FF FE FE FE FD " + ;
"FF BE D2 CD AB BF C0 A7 A8 A6 B9 9F 98 D6 D5 BB " + ;
"C9 CA C8 44 44 52 00 05 07 76 9D 95 79 8C 7B A3 " + ;
"93 7C A0 A7 B0 B3 B0 A8 C7 AC 91 F7 D3 AB C7 AE " + ;
"86 CA BA 9D D7 C7 B6 EC E3 D6 F6 E0 D4 E7 ED D6 " + ;
"B7 C4 B4 C6 F7 F5 F8 F8 FF DC F2 F8 C4 E1 E5 F6 " + ;
"FB FF 9B BF BF 72 6A 71 7B 71 53 E9 DE E6 AB 86 " + ;
"B4 60 40 23 6D 45 33 AB 81 6E 9F 73 54 B0 8D 6B " + ;
"D9 BF 9A CA B0 8B E9 C9 A6 EB C5 A7 D7 B5 9D E9 " + ;
"CA B3 FF F4 CE 71 5D 44 69 4C 3D A3 7B 69 C3 95 " + ;
"76 EB B8 97 9F 6B 54 74 40 33 C2 9A 70 E9 CF A7 " + ;
"DD B6 89 F1 BB 92 CD AE 79 E7 C4 C1 90 64 63 A6 " + ;
"83 62 92 6A 58 61 57 39 A3 A0 91 B5 B2 AE DD EA " + ;
"DC E5 E9 E4 95 7F 8B 4B 36 35 EA CD C8 9F 91 92 " + ;
"66 4C 3C E4 D6 D0 3E 42 4D 00 01 10 8E B2 C4 94 " + ;
"8C 97 92 9D 9B 51 6B 65 BD C8 B8 E2 D9 D0 C5 B7 " + ;
"B9 F3 F0 E2 A8 9F 8B E8 E6 E6 A5 B1 B1 38 45 3D " + ;
"E0 DA CF 80 77 6D 39 45 3F F4 FF FF FF FC FF FC " + ;
"FB FF FD FF FF FD FF FF FD FF FF FD FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FD FF FC F8 FF FF F6 FF FB FF " + ;
"FE FF FF F9 FC DB EA E6 E1 EA F3 4D 36 3E 81 60 " + ;
"4C A8 7B 59 3E 0F 01 E4 D7 E5 6F A8 A5 63 BB AB " + ;
"5E 90 86 B5 BC A1 DE CF B5 D0 C9 AE D8 CA B8 BE " + ;
"AE A8 D9 DB DC D7 DC DF B9 C0 BB A7 BA A9 61 76 " + ;
"5D 45 4A 2A AF BB 9F AF E1 D7 EC EC FA FF FB FF " + ;
"F2 FC FC FF FE FF E1 D5 F3 61 48 4C 48 43 12 C1 " + ;
"CB AD 76 67 83 34 19 00 70 4D 25 A0 82 65 EE DB " + ;
"B8 D5 BE 98 D0 B3 87 F0 C6 9B F3 C0 98 FC C8 A4 " + ;
"DA AB 8F ED C5 AC E4 C4 A7 FF F9 E2 7E 63 4F 99 " + ;
"79 62 59 31 14 CD A0 7F EA BB 9F A2 76 5F D0 9C " + ;
"74 CC B0 87 CD A9 7B F0 C1 9B C7 B7 89 E9 D3 D8 " + ;
"61 3F 46 85 70 55 9B 9A 8C E0 FB EB D0 F0 EB BA " + ;
"CF C7 AA BC AB F2 FB FE C4 AE C1 64 33 31 77 4E " + ;
"38 C5 A8 8D 64 3D 11 74 55 40 CB D3 E0 5A 8C 82 " + ;
"93 C6 AA 7A 77 79 71 82 7E D8 E7 DF E2 D7 C1 D4 " + ;
"AC 99 C9 AE A4 F3 E5 CF CB C5 AE 78 73 75 B4 B8 " + ;
"C3 F4 FF FF 6D 73 6E F3 FD F7 C6 E8 E7 C8 E5 EA " + ;
"F7 F9 FF FD FC FF FD FF FF FD FF FF FD FF FF FD " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FE FA FB FE FF " + ;
"F2 FF F7 FF FD FF FF FD FF F3 FF FE AF B8 C2 BD " + ;
"A1 AE 74 5C 4A 95 79 50 9B 81 5C FF FB F7 FF FF " + ;
"F8 FD FF F8 EF F9 FF C6 F0 DF E2 D7 E1 B7 BE C1 " + ;
"D3 DB DA 99 9F 9E 56 61 5F 72 6C 67 8F 75 67 9B " + ;
"87 6E DA D2 AA FB DD BA EE BB AB DF EA E7 C8 FF " + ;
"FA E3 FF F3 E6 FA EE C9 EA E6 F8 FB FF E5 F3 F9 " + ;
"98 9D 9B E5 DF D2 88 6B 92 80 5F 63 90 6E 40 6E " + ;
"47 20 BE 94 69 BF 9B 73 D0 B0 8D E7 CB AD D9 BE " + ;
"A4 E0 C2 A9 ED CB B3 F9 D6 BC F2 C0 AC FF E2 CC " + ;
"DD BF A6 4F 34 19 5A 38 1B 97 6F 52 C5 9E 7E DF " + ;
"BC 9A DD 9D 7A AD 87 64 E0 B2 89 CB 98 78 FC F3 " + ;
"CD BD B6 C3 A0 93 A1 D0 DD C5 CB DC D9 A5 95 AD " + ;
"9F 85 95 BC A4 A4 97 80 8E 7B 78 87 DF E9 E9 B4 " + ;
"AD B4 56 61 41 F5 FF FB 78 90 7E D5 D3 BB A7 A8 " + ;
"C4 3E 5C 7F BF D0 CC 96 9E A5 BA C3 C0 AA 9A 93 " + ;
"D0 AD 93 E8 BC 9F C3 BF A7 F3 ED D0 C8 BE A6 99 " + ;
"8A 8E 4E 27 19 76 56 3F 67 42 28 5D 44 2A A3 AC " + ;
"98 ED FB F0 FF FF FB FA FC F6 FD FF FF FD FF FF " + ;
"FD FF FF FD FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FC FF FF FE FF FF FF FE FD FF FD FD FF FB FD FF " + ;
"E7 EB F6 CA D0 DB 84 6A 5C 88 5C 4B FF FC FF 63 " + ;
"78 80 A7 CE D0 B9 C8 DB 8F 86 A1 B2 C5 CC 8F 91 " + ;
"8B C7 D3 BF A8 A9 95 C2 C4 AE D8 F6 D3 E8 FE E5 " + ;
"F5 F0 E7 FF FE F3 F6 EF EC E5 D5 CF EC E2 D1 CD " + ;
"D4 C7 DF EF E8 AE BE AC FD FF FC 7C 79 A0 9A 7D " + ;
"4A AA 77 7B 7F 47 54 96 9A 8F FF FF FC 5C 49 41 " + ;
"40 48 23 68 51 31 C5 7F 6E DB 89 77 CC A9 7E CE " + ;
"AB 91 BC AB 90 AB 95 6C B5 97 7E A9 90 7C B5 9F " + ;
"7C C1 B7 95 F9 EC D2 C3 9A 81 6A 4A 3D 43 36 20 " + ;
"E6 C2 8C FF D1 A5 DF BA 80 EE B0 87 A8 6F 56 E1 " + ;
"CD C2 E0 E5 E4 A7 A9 B1 C2 CA D1 69 82 86 B8 B6 " + ;
"AC AE 8A 90 B7 BD AC 9E A0 A8 A3 B1 99 A1 C6 AC " + ;
"99 9C A1 7C 79 71 E6 EB EE 73 7B 7A 93 83 7C 9A " + ;
"7D 6E C0 C5 AA 13 26 2E 52 6C 66 E5 FF FE A6 A7 " + ;
"97 9F A0 90 BC A1 8D A3 8A 76 C5 CB BA DF D5 C4 " + ;
"CD B2 A4 8E 92 8D 75 58 51 90 79 49 95 82 57 62 " + ;
"4B 25 80 74 70 EF FF F9 F6 FF F7 F7 F6 FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FE FF FF FE FF FF FF FF FC FE FE FD " + ;
"FF FB FE FF F2 F8 FD E0 E8 EF 7C 7D 6D 75 4E 3F " + ;
"A8 6A 64 57 3F 23 84 8A 79 98 97 A1 0A 05 00 FF " + ;
"F9 E3 FA E8 D1 EE CC B5 F0 D4 B6 EC CB AA FB C4 " + ;
"A9 EE BF A3 D5 C1 A2 EA D2 BA EC D2 B4 E7 CF BB " + ;
"E4 BE A0 F1 AF 96 FF EE EA CC D0 CA E3 BA C1 D3 " + ;
"9A C7 56 40 24 6C 46 3A D7 CF C2 FF FF FB B7 B5 " + ;
"C1 3A 43 47 2D 2D 1F 2A 59 39 49 55 37 AE 87 71 " + ;
"F8 C7 99 FA BC 98 E1 C6 A4 E9 D7 A2 FB D4 AD E5 " + ;
"B2 9E B9 8F 83 A3 7E 68 A6 92 73 ED C3 A6 5E 43 " + ;
"39 0E 07 00 7C 5A 2F F5 D9 B6 FF B7 88 C4 8E 6B " + ;
"C1 B8 A4 9F C4 BC 7A A0 A4 BB C6 CE BD BC C0 AF " + ;
"B8 B5 9B B6 B2 D4 D0 CF C8 C8 CE CD CF D7 AD B9 " + ;
"A3 C6 D1 CF FF FA FF 4C 18 22 7A 54 32 82 65 4A " + ;
"98 6A 59 6A 3A 2E D7 CD BC EB F2 FF C6 D3 CB F2 " + ;
"FF F4 B7 E4 E7 91 C0 B8 C1 E8 D9 9D CD C1 C9 ED " + ;
"E7 D2 C3 BA DF BB B5 92 78 85 7E 52 4B A4 7D 5D " + ;
"A2 7E 60 6B 4E 29 A9 91 91 FF FD FC FF FF FB FF " + ;
"F9 FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FE FE FE FF FF FF FF FF FF FF " + ;
"FE FE FD FD FD FD FF FF FA FF FF F5 FD FC C3 D7 " + ;
"CB 5F 51 3F B0 6F 5A DA A0 83 7A 62 3E F4 E8 E2 " + ;
"09 02 07 61 53 29 F8 F1 D0 DB BA A6 E6 C6 A9 F8 " + ;
"D2 B0 FF C9 B3 DD AE 92 E8 D4 AB EA D1 B1 C7 B9 " + ;
"88 F8 BF AA DC C3 97 E4 BB 8A FF D4 C7 E3 FB F3 " + ;
"A3 BB B3 E3 D5 E7 65 5C 4F E4 E2 D7 FB F7 F2 7F " + ;
"43 43 A5 79 72 C0 DE D9 00 08 06 48 2D 18 61 85 " + ;
"75 7D 8B 85 BC C3 9E B8 94 76 DA A5 90 EE B9 86 " + ;
"F8 CB 92 EA C8 9D F6 D7 A0 FF DD B0 F8 DF BD FF " + ;
"E2 CC 3D 2B 20 6C 6F 55 9A 7E 55 3D 36 1B DA D4 " + ;
"BD C5 BD B0 EB ED ED 9D AC B5 83 8C 9A B9 B5 C1 " + ;
"C6 CB CE BA D5 CC AA C2 C8 AF C2 B9 A2 A7 BC AB " + ;
"C1 C6 9A B0 9D CE D1 DF E1 E8 EB 65 48 58 7C 4E " + ;
"25 9C 7C 59 AA 80 69 80 55 46 CA C4 B1 7E 80 92 " + ;
"02 01 00 FF FF FB 82 90 AD B7 BE C1 9C AE A7 AE " + ;
"D4 D8 F5 F6 FA D6 B5 A6 E1 DA C7 85 82 84 84 5C " + ;
"49 91 6A 5B 97 79 60 4C 4D 21 CA D0 CB F5 F8 FF " + ;
"F8 FF FE FD FE FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FB FD FE FE FD FF " + ;
"FF FF FF FF FF FE FD FE FA FB FF FA FB FF FB FA " + ;
"FF FB D6 EA DE C5 D0 CD 53 45 2E 69 43 31 6A 4F " + ;
"3B C2 BD BA B4 BC DA 00 03 00 8C 97 83 D7 DB CF " + ;
"B3 9D 91 C0 A9 9A DF E2 D2 C6 CA B1 B6 A1 85 E1 " + ;
"C6 AC FE E2 C0 FC C9 B9 EA D0 AC E3 C6 93 D6 BA " + ;
"9C FF FF F7 BF D0 C3 BA F4 DB EA FD FF 8E 71 6A " + ;
"7C 51 42 B1 7B 64 4F 40 20 FF FC FF 7A 3A 50 36 " + ;
"21 02 46 48 35 06 0F 02 A3 C5 A7 C3 CB C1 C9 AD " + ;
"AD B7 9F 81 B6 A1 81 DA AC 94 ED B2 81 E8 AE 7D " + ;
"D4 A8 73 E9 C2 9B FF E7 D3 51 41 24 D7 B8 91 AB " + ;
"A0 8A 6E 8E 89 C5 DD DB 60 77 79 6E 93 97 83 B3 " + ;
"B9 53 7F 86 72 92 98 80 97 9F 9E B0 B7 83 9D 90 " + ;
"A7 AF BC 71 97 9B CF F3 E3 D6 E2 E2 CC EE ED F8 " + ;
"FC FF B6 AF A6 49 43 36 60 3D 2F 8C 5A 46 A2 97 " + ;
"7B EE FA FF AA CA C5 C2 EA E9 B4 B1 C0 C0 AD 98 " + ;
"F2 F5 DC 8E A5 A1 E9 C8 C5 C6 98 80 FF FF EC 67 " + ;
"59 5B 8A 73 46 8B 66 58 98 72 5A 52 56 25 F3 FF " + ;
"F6 F8 FA FF F6 FF FF FC FE FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FC FE " + ;
"FF FF FE FF FF FF FF FF FF FE FF FF FC FC FE F8 " + ;
"F9 FE F5 F9 FF F7 FB FF F2 D4 DC ED 7F 92 8F A6 " + ;
"A7 93 FF F9 FE EA FF FF D7 F4 F9 31 3D 67 00 03 " + ;
"00 F3 FB F1 FF FA F9 DA C5 C4 F5 FA EB FF FF EB " + ;
"F8 DA C9 F4 D2 BA E1 CF BE EA CF BB F2 DA C2 E5 " + ;
"C7 A4 DB B5 92 F2 CD C5 FD EF E9 DF FC D5 F7 E5 " + ;
"FC 6D 50 42 8E 86 69 70 5C 43 DF DC C0 95 8E 93 " + ;
"F4 C3 E3 1A 23 26 4C 31 1C 39 17 00 6E 67 4C E7 " + ;
"DF E0 CE AA AA FD FF F0 B5 D2 C3 F5 D5 C8 CB B7 " + ;
"B2 B9 B7 B6 E8 DB D3 2F 2B 26 C5 C9 CE F7 EC E8 " + ;
"3B 32 25 FF FC FC 3D 48 38 B6 C1 B1 FA FF F7 D0 " + ;
"E0 D5 C5 E2 D9 C1 E5 DF C8 E2 E2 DA E5 E9 E5 F3 " + ;
"F2 ED F3 E8 F3 E8 E0 DC E7 EF E3 F0 E8 BB BC A2 " + ;
"EA F1 F4 6E 54 5A E5 DC D9 FF FF F7 DD CD C0 64 " + ;
"4F 39 85 92 72 8F A2 A7 1E 2F 26 F4 FD FA BE A1 " + ;
"9A AC 82 57 CE CF A7 E2 FC EF E6 CB BD A9 8F 77 " + ;
"F7 FF F5 70 42 4F 7B 78 3B 94 76 63 73 4B 2E 85 " + ;
"7F 56 FB FF F9 F4 F2 FE F6 FC FB FF FD FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FD FF FF FD FF FF FE FE FE FF FF FE FF FF " + ;
"FE FF FF FB FD FF F9 FD FF FB FF FF F8 F8 F4 FA " + ;
"CB D5 DC B5 B9 BA D6 EB E2 6F 98 89 D5 E1 E1 FF " + ;
"F9 FF 1E 1E 1E 16 0F 06 FF FF F7 E5 DC D3 CC A7 " + ;
"91 E1 B8 A1 F2 DA C4 EF D6 B4 DE D1 C1 CE B5 95 " + ;
"CF BA A4 F5 DC C8 CF A9 87 BE B0 9E EC F5 F2 BC " + ;
"B7 9E FF F8 FF CD B2 BB E0 E5 E8 DB DE D5 CD C9 " + ;
"B7 6A 67 69 EB CF EC 52 44 5C 53 26 19 53 21 03 " + ;
"60 53 2D 88 7B 6D B7 69 58 ED CE AF 9F B9 AB AC " + ;
"A4 93 FF FF EE C2 E1 E4 D8 DB E3 EC F0 F5 51 5A " + ;
"5E EC E5 DC A6 AA 9F 4D 3F 40 CE C3 A8 10 13 00 " + ;
"9E 9D 88 E3 C9 B8 E2 B9 AA EF CC B8 AA 8F 75 BA " + ;
"9E 80 AD 8A 87 C4 A3 9A 9B 7C 6D DB BC C3 88 63 " + ;
"5F A2 8F 6A DB C5 CA 79 49 48 62 48 2A 78 6F 54 " + ;
"D7 C6 B3 F6 DF CF E9 EF D8 CE CD D7 5B 4A 3D FF " + ;
"EE DF C2 B5 B7 A2 7B 5F C7 C6 B2 BE C9 C6 E1 C1 " + ;
"AE B8 B2 9F E6 FC FA 41 2A 28 A4 84 53 AD 86 78 " + ;
"4C 31 16 A6 A9 93 DE F2 ED EF FF FF F8 FF FE FF " + ;
"F9 FE FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FB FF FF FD FF FF FF FF FF FF " + ;
"FE FE FF FE FF FF FF FF FE FF FD FD FF FE FB FA " + ;
"FC FF FF FB F8 F7 F9 E1 E4 EC 91 A2 97 FA FF FC " + ;
"F6 DD E1 65 3F 1C FA F3 F8 04 01 00 00 09 00 D5 " + ;
"E2 CA CE B0 95 EB C0 A5 D9 BD 9F E9 CD AE FF D5 " + ;
"C8 AA 94 6B C9 AB 92 D6 A9 9E FA F0 CE EF EA CB " + ;
"DE C4 BD E1 ED DB D5 CB B3 BE A1 B0 9C 9E A9 C7 " + ;
"CA C1 A7 A7 A1 FF FC FF 69 3B 41 DF CF C8 27 01 " + ;
"00 6E 3D 23 38 35 09 2E 44 28 87 44 2F B6 7B 61 " + ;
"FF F7 EE 9B 7B 76 EB 9D 6F F3 DC C2 D6 B4 89 EE " + ;
"C6 95 E7 C7 A4 79 5D 3F BA AF 94 B2 8F 75 D0 A9 " + ;
"9A FF FF F0 10 29 0F 9A A7 8D E5 E4 CA A8 A1 86 " + ;
"D2 BA 9E CE 9F 84 C1 97 92 93 83 73 F9 FB F5 FC " + ;
"DA D4 DE AA 9E CB C2 A7 B2 A4 A5 81 55 4E 95 7F " + ;
"5C 7B 64 4A 86 52 42 70 2D 24 D3 B6 A8 A5 9E AD " + ;
"C0 CB C1 F1 FF F5 8F BF C0 DC D7 C8 C5 D6 D2 EF " + ;
"F6 F3 B2 92 75 FF FF F1 F3 FF FF 60 63 4A 4B 17 " + ;
"07 4A 24 24 73 74 60 CB D1 D8 C0 CC D2 ED FA F8 " + ;
"FF FF FB FF FC FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FA FF FE FD FF FF " + ;
"FF FF FF FF FE FF FF FE FF FF FE FF FF FC FE FF " + ;
"FE FF F7 FF F5 F8 FD FF FB FF FF A3 B9 9C 97 8F " + ;
"A6 41 0E 3A 99 71 54 A7 73 4B C4 A7 B6 D7 D6 E0 " + ;
"00 05 00 47 54 3A FB FF EC B8 A3 88 F5 BF A6 EB " + ;
"B3 A2 EA CC B9 D9 B2 8C EC EA CB F1 F4 E4 DA CD " + ;
"AD CF AB 83 E0 AE 9A F2 FF F1 B2 80 38 E1 DA C7 " + ;
"92 A1 A3 AB C2 BD FF EA E9 34 19 05 9A B7 90 FB " + ;
"FF F8 B4 A9 B1 2F 00 03 5D 44 34 38 4F 41 64 30 " + ;
"29 73 3E 23 AF AB 93 EB E2 CE E3 AF 81 AC AC A6 " + ;
"ED E7 DA BC A1 87 ED D8 C2 B5 B5 9D BE C6 B5 C9 " + ;
"B5 A3 44 2F 20 FF F5 E3 B4 A0 8E 5E 3D 2D C7 BF " + ;
"B2 E1 FF FC B9 EE EB AA CE D6 B5 D1 BE B6 EA CC " + ;
"80 BA BF C2 BA A3 E3 C4 AB 5D 73 67 B3 AC A9 68 " + ;
"29 25 A8 72 72 90 6D 69 97 74 6A 73 57 46 BA C9 " + ;
"AE A0 B7 C6 70 83 88 F2 FB FF B7 AD A3 D9 9F 8C " + ;
"C4 B7 AF E9 E6 D8 C1 A4 7F FF FF F7 69 63 76 BE " + ;
"C4 AD F8 FD FE 45 5B 60 B4 EB DC CA DF FA C9 D0 " + ;
"E1 F6 FF FE FF FF F6 FA FE FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FD FF F8 FF FF FB FF FB FE DA " + ;
"DB DF C3 CA C7 95 92 83 76 58 3F C8 91 72 87 6D " + ;
"3F F3 DD D8 A7 A8 BC 00 00 04 79 8C 79 E2 FA E2 " + ;
"D1 BA AB F8 BE B2 E9 D7 D6 F5 FA DA D2 C5 B5 DB " + ;
"9D 75 F8 BE 8D EA C3 A3 DB AD 7E F4 D5 BE ED FF " + ;
"EF EB E6 C7 DE CC C1 7A 6F 6B C4 C1 BC B5 C1 C5 " + ;
"FC FA FA 89 52 55 F9 DF CF AD A6 97 40 1C 0C 4D " + ;
"4E 44 8E 66 4A 71 41 07 52 3A 1E 8F 9A 9E C4 8E " + ;
"70 DB A8 88 DC BF A4 D7 DE C9 BF B1 9E CC 9D 82 " + ;
"7F 66 4C BD BD AD 91 8B 8C 5C 58 4D FF F9 D9 B7 " + ;
"7E 5F 75 45 29 B6 79 57 DD A4 85 DA CA BE F7 FE " + ;
"F9 F8 D8 CB A8 9E AF B0 A8 B2 5C 75 53 81 B8 9D " + ;
"E9 F1 F1 9B 97 A2 58 33 17 84 69 4F 8E 68 50 75 " + ;
"4D 3A D9 DA C6 DE FF F2 94 C3 C0 C5 F3 FF A6 98 " + ;
"9C C4 A7 92 F7 FE F7 BB B4 A0 F0 F6 E5 B0 94 A7 " + ;
"73 30 27 45 3D 1F 91 6F 41 E0 CF B5 FD FF FE A3 " + ;
"AF B5 F3 FA FD FF FF FE FF FF FF FB FD FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FC FF FD FF FE FF " + ;
"FF FD FF F9 FC FF EB F7 F7 B4 BD B3 86 7C 6B 95 " + ;
"77 64 AE 95 73 63 42 2E E4 CB C7 FF F7 F9 00 04 " + ;
"00 15 29 12 FA FF F0 C8 CF C0 C4 A2 77 B4 B0 9D " + ;
"EF EF E1 C0 B3 B1 BA 99 8A F2 C5 9F F0 C6 A3 DD " + ;
"A7 7E DE E5 C0 E7 CD C6 CF C3 C3 FB FF F6 89 86 " + ;
"78 FF FF FB FF FF FC 91 5A 53 3F 3A 2B FA FF F9 " + ;
"BB A6 A8 11 1C 2A 8F 68 6A 91 54 34 68 33 1E 52 " + ;
"3C 31 20 39 13 7C 61 3C B5 78 58 C4 94 7C F5 F9 " + ;
"DC B5 C3 9F C5 B1 92 71 59 43 FF DB C5 8A 6E 50 " + ;
"97 8D 75 D0 D1 C8 70 6D 65 56 67 52 BA D2 BA E2 " + ;
"DB CC AB 8D 82 A2 7C 5C AC 81 6C 6E 53 39 61 5D " + ;
"34 F8 FD E8 AB A9 AF FF F8 FF FD FF F7 98 66 66 " + ;
"74 5A 4C 6A 81 6B B5 AB A4 99 76 68 FD FF F2 C2 " + ;
"F6 FD B1 BD A7 DD E7 E1 F8 CF B8 9D 98 7F C2 CD " + ;
"C5 30 15 07 98 63 4F A4 76 57 86 66 3D 29 18 03 " + ;
"CC CE CE D8 E2 E9 FA FF FF FF FE FA FF FF FE F4 " + ;
"F7 FB FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FB FB " + ;
"FF FF FA FF FF FC FF FB FE FF EA F9 FB EB FE FB " + ;
"BB C5 BF 42 3F 3A 99 5E 5C CB 95 74 7E 56 2C BE " + ;
"B7 A8 62 67 70 04 01 03 3B 42 35 B8 CB B8 E9 D6 " + ;
"B5 F6 E5 E8 D3 CA C0 E3 F3 FF D3 CB CB E5 AA 83 " + ;
"F0 BA 97 F0 CD 83 DD A9 7A FB FD FF F2 F0 E8 AB " + ;
"96 7A DB E4 D7 6A 6C 6C AA 81 7E 7F 5C 3B B5 BB " + ;
"A8 D2 C3 C0 E6 C4 CA AC CB E0 00 03 10 5A 34 22 " + ;
"94 51 3C 82 53 34 34 4E 37 2F 17 00 75 3E 1F A3 " + ;
"63 4B 8C 6A 53 DF D8 BF DC BB A7 A1 66 57 9C 77 " + ;
"5D FF EF C9 3B 30 1C CB F0 EE 6B 8B 86 E6 FC F0 " + ;
"34 28 22 80 45 43 AD 8B 6D 9C 89 5C B3 80 55 51 " + ;
"3B 18 FC FF E6 C4 C2 B8 B2 BA C1 14 01 04 40 45 " + ;
"3C FF FD FF FF FF F5 76 7B 6C D1 D4 D8 91 77 69 " + ;
"9C 83 6F EE F6 FF 46 5B 42 F4 F1 F3 AF 8A 74 DB " + ;
"E6 D6 9E 86 86 86 53 39 84 6E 52 AA 6A 51 74 57 " + ;
"3C 66 56 49 BD BF C0 F3 FB FF FA FF FF FF FF F9 " + ;
"FF FF FB FD FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FB FE FF FF FD FF FF FB FE FD FD FD F6 FF " + ;
"FF E5 F7 F6 CF DE E0 BE C9 CD 71 44 47 73 56 37 " + ;
"8A 73 45 AA AD 94 F9 FA FF A5 96 AB 01 00 04 49 " + ;
"4D 42 F9 FA FE CF C3 C9 C2 AD AB FF ED E2 D3 BB " + ;
"A3 EB BD 94 F3 C5 95 CC C6 83 FB B4 93 A8 A3 8A " + ;
"E0 D9 C0 DD C0 B7 FD FF FC B4 AF A0 76 50 3E 6A " + ;
"5C 40 EA FA E8 B2 8A 85 83 3A 32 D0 C7 C4 F0 FF " + ;
"FB 5B 6B 53 3C 27 12 4D 37 13 59 5F 5A 39 45 33 " + ;
"3D 39 20 4F 39 20 57 37 24 57 51 3E 81 9D 89 A0 " + ;
"B2 A1 5F 60 56 A9 9E 83 BD AF 93 6D 76 61 7A 73 " + ;
"5A 64 59 3E A0 91 81 64 3E 3A 71 56 2A 6B 6E 4E " + ;
"46 38 22 B9 B9 B9 AE B7 BB 77 69 6A F5 F0 F2 F7 " + ;
"EE E1 8B 7B 6B 91 95 90 CD CF C3 AC 9A 8F FF EC " + ;
"EE C2 C5 B6 CE D3 BE D0 BF CC D0 E8 E0 F1 DC CD " + ;
"9D 95 96 F0 EC F2 63 4D 41 87 66 52 86 75 5A 85 " + ;
"5F 47 29 12 02 E6 DA D4 D6 D7 DB F4 FA FF F9 FE " + ;
"FD FF FF FB FF FF FB FA FD FB FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FA FF FF FA FB F9 FF FF FC FF " + ;
"FD F8 F7 F9 F3 FA FF FE F4 FE FE E2 E9 EC A4 B9 " + ;
"A3 87 A6 97 E9 F2 E8 FF FF F5 BC B8 B3 B9 B8 C2 " + ;
"C5 D7 E2 00 02 06 00 04 07 DF E9 DD FF FC FE 99 " + ;
"71 58 F6 C8 99 E6 C4 99 DB B7 81 DE C2 A4 D7 BF " + ;
"91 C7 89 65 AD 8D 8E F8 FF FA C7 B5 96 C3 A9 99 " + ;
"FE FB F3 68 5D 47 FF FC FF 73 55 54 BB 8C 70 91 " + ;
"65 46 7F 67 51 C5 DC C6 F6 FF FF 61 65 59 50 17 " + ;
"16 8F 82 7A 53 48 3A 64 57 3D 52 50 31 3E 49 29 " + ;
"1C 2C 14 41 42 32 46 44 3A 3A 38 20 63 53 2E 8F " + ;
"4F 30 90 4D 32 9B 61 3E 68 39 1A 76 53 49 5A 41 " + ;
"21 02 03 00 E0 EC F0 E1 D5 E9 50 2B 33 9B 69 5D " + ;
"86 58 51 53 4D 36 BA AD 93 FF FC F6 5C 5F 56 F6 " + ;
"FF EE 5A 45 36 99 7B 70 FD FF F0 AD CD BA FD FF " + ;
"F9 91 6A 54 FF FD FF 75 58 68 6A 44 32 82 61 47 " + ;
"93 64 4F 51 31 14 BD AF A9 CD C5 C5 ED EE F2 FB " + ;
"FE FF F7 F9 F9 FD FF F9 FF FF FB FF FF FC FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF F7 FD F8 FF FF FC " + ;
"FF FF FB FF FF FB FF FF FB FF FD F9 FB FC FA FF " + ;
"FF FF B2 C3 AE E3 FF FF 47 5D 68 4B 53 4C 7D 7D " + ;
"71 CE D0 D1 FB FD FF FF F1 F2 00 05 00 01 09 00 " + ;
"B4 C6 C5 DB F1 D9 BB A0 7E FF C3 A6 FF C4 94 F0 " + ;
"B4 90 F5 D5 B2 C2 AE 8B F2 E4 DE E1 C8 BE FF E1 " + ;
"CD F1 EA D7 C5 D2 BA FF FE F7 AB 9E AC 55 40 49 " + ;
"AC 94 76 BA 89 63 94 6A 53 D9 DD D8 D0 C2 DA FF " + ;
"F2 FF D1 A5 98 51 32 29 64 51 44 40 17 00 70 3F " + ;
"1F 65 37 18 70 2F 1A 64 1C 0A 64 65 45 5B 70 4A " + ;
"3D 49 1F 82 55 3A 28 0C 00 5B 33 17 59 2B 0C 09 " + ;
"05 00 45 56 41 FF FF FB B6 A9 B1 52 20 2A B4 84 " + ;
"6C A1 81 58 A4 73 5D 91 7A 64 7C 63 39 62 64 50 " + ;
"E8 F6 EA BD C2 AD D6 D9 C3 E2 E3 D9 E1 DF D5 FF " + ;
"FF F0 91 6F 58 9A 98 8D 9E BA A6 E7 EE F1 44 10 " + ;
"0A 81 44 24 8A 53 44 45 49 1F FF F8 FB A5 A2 A4 " + ;
"F4 F3 F5 FD FF FF FA FD FB FC FD FB FF FD FC FF " + ;
"FF FE FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FD FF " + ;
"FE FC FD FB FD FB FB FF FE FF FF FE FF FF FC FD " + ;
"FF FE FD FA FB F9 FF FC FF AC A8 AE 95 A8 AB AF " + ;
"C0 BD C6 C8 C2 DB CF C5 7A 56 4C DD AE A0 F7 FF " + ;
"FD 25 22 24 00 04 00 45 77 6D D5 D5 C9 FB C6 B1 " + ;
"E9 B3 90 D7 A3 74 E9 C4 A8 D5 ED EB E9 F0 EB A0 " + ;
"78 5C D0 B8 A2 FF FF F5 B1 A3 90 F4 E4 D3 CE E5 " + ;
"E0 B3 95 A0 84 5C 4A A4 72 56 9E 7E 6B D8 D4 D3 " + ;
"D1 A8 BD 55 2C 29 F9 FF F5 E0 E1 DD A5 BA B1 93 " + ;
"9C 87 78 61 4B 6D 4B 3B 62 31 21 6D 37 1E 2C 18 " + ;
"00 68 47 2D 3B 25 0C 36 25 1C 15 06 03 2F 28 19 " + ;
"80 8B 77 B6 CB C2 C5 D6 C8 A2 97 81 EF D0 CF 88 " + ;
"5B 58 88 6A 47 8E 7E 5A 99 7A 63 93 73 56 66 46 " + ;
"1B D7 F1 DA B9 D8 C9 44 37 27 D8 D0 C9 D5 FF F7 " + ;
"EB FF F2 9E 5C 5E 98 9A 84 DA FE FE 3E 47 44 EF " + ;
"FF FE 6A 81 83 73 7A 6B 1C 14 03 D0 DF C3 A2 A2 " + ;
"A8 E4 E5 E9 FF FF FF FB FC FA FF FF FE FB FB FB " + ;
"FF FE FF FF FC FE FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FD FD FD FF FE FF FD FD FF FA F9 FF FB F8 " + ;
"FF FF FE FF FD FF FF FA FF FD FA FE FF FF FA F9 " + ;
"E9 F5 F5 B7 C0 C9 76 72 6D 33 1B 00 AC 7F 5D 82 " + ;
"62 4B 75 45 3F FC FC FF C8 E0 E0 00 00 06 0A 01 " + ;
"00 A1 B9 99 CB DB C4 F7 D8 C3 E1 EF DD E8 DA E5 " + ;
"D5 D8 DC E1 ED E1 C0 B0 8C C3 A1 7D FC E4 D8 FF " + ;
"FA E8 A5 86 6F FF E6 E6 E5 EA D5 69 60 45 7B 61 " + ;
"51 C4 A7 A2 77 3D 3E BC A3 79 A8 A4 AF D1 C4 D4 " + ;
"A2 7E 8E 96 88 89 E5 ED EC FB FE FF F6 FE F7 E8 " + ;
"F4 D8 DC CE D2 E2 DB CA DE D2 C6 69 77 75 FA FF " + ;
"FB DE F4 E8 42 4E 52 30 16 26 50 17 0E 60 49 23 " + ;
"4F 44 3C FF FD FF 68 3F 36 92 53 55 98 77 68 4E " + ;
"3B 10 D2 DD BD F4 E9 EB F8 F9 F7 8E A5 97 F5 FC " + ;
"FF FF FE F8 64 5A 42 AD B1 A6 F7 D8 DF 84 7B 85 " + ;
"84 8D AF 30 3F 48 D2 D8 EB E0 CE F3 DA E6 DA A9 " + ;
"B5 B5 F2 F5 FD FB FE FF FF FF FF FF FE FD FF FF " + ;
"FF FC FB FD FF FD FF FF FB FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF F8 FF " + ;
"F9 FF FF FF FF FB FF DA DE E9 A5 B5 BB 82 90 8C " + ;
"4B 44 35 87 6F 59 B4 87 65 66 43 18 FB E7 D6 FF " + ;
"FD FE 00 02 00 00 03 04 11 22 2B FF FF FA B6 D1 " + ;
"C7 FF FE FF D8 CA CC EC DD CA D8 AF 82 F9 C5 90 " + ;
"C9 AF 87 EC E9 D4 FF FA E0 8E 7F 6F E7 D6 D3 FF " + ;
"FC FF 8E A3 A5 89 8B 8B 59 46 3F A1 91 81 D5 CE " + ;
"C5 97 7D 77 B6 8A 79 85 5E 3E 82 76 54 D6 DB C6 " + ;
"5C 52 41 9A 75 5F C3 A3 9E 98 8C 70 BD B5 B6 4E " + ;
"46 57 DC DF CF E8 E8 E2 CD C8 D7 BA C1 B4 68 65 " + ;
"39 66 42 1A 67 33 1C 69 80 72 FB FF F9 66 45 36 " + ;
"4E 3D 30 CE E1 DE FF FF F7 8B 70 56 B8 D3 B2 E6 " + ;
"FE FE FF FF FA 5F 36 27 83 73 74 FF FD FA 53 1F " + ;
"18 70 47 38 C9 B3 A1 C7 C4 BF 77 83 8F 12 20 36 " + ;
"D0 D7 E6 C9 C8 CA FC FE FE FC FE FE FC FE FE FC " + ;
"FE FE FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF F8 FF FA FF FF FF FF FA FF FB FD FF EA F6 " + ;
"FA D5 E0 DD 93 8F 84 65 52 43 82 4A 37 B5 8C 76 " + ;
"7A 50 33 78 5E 46 F9 FB FC C1 BF CB 0F 03 09 00 " + ;
"00 06 1D 35 2D A0 AF AB FA FC F6 C3 A5 92 E9 AE " + ;
"87 FC CA 95 E1 C2 8D D4 B4 89 E5 C9 B1 E4 ED D2 " + ;
"B4 D1 B7 C7 AC 9E FF FF F3 FB F7 F2 E8 FF FF A5 " + ;
"C4 C1 75 65 59 8B 5F 58 BC 7E 73 B8 88 70 81 69 " + ;
"55 CF C3 B9 85 70 61 9A 7B 5C 7C 5E 59 9B 92 71 " + ;
"EA EB D7 FB FC FA 19 1C 13 AE AD B1 91 92 9C 1F " + ;
"2E 19 55 8E 61 7E 55 3E 7E 4F 39 46 2A 12 7C 60 " + ;
"55 EC F7 F5 DE E0 DA D9 C0 AC C9 9C 8F A9 B5 9F " + ;
"FF FE F8 EA D1 D5 52 49 2E B2 DB BF D6 CB D5 57 " + ;
"2A 1D 7C 5C 3F 80 61 40 63 47 25 2C 18 06 C0 B8 " + ;
"B8 DF E2 EA B1 BC BA F6 FF F8 FC FE FE FC FE FE " + ;
"FC FE FE FC FE FE FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF F9 FF FD FF FF FE FF FE FF FA " + ;
"F7 F9 F6 FE FD F6 FF FE D0 D1 CD C5 BA B6 5E 61 " + ;
"52 65 4B 44 55 4E 2D C3 D0 AA CE C7 C4 E6 E2 E8 " + ;
"F8 FF FF 76 69 79 00 02 00 02 00 00 37 4C 44 CF " + ;
"EF DE E8 E1 C8 D7 9E 7E E2 A2 79 F7 CA 97 F3 BB " + ;
"8A D1 AA 84 E1 FA DE F7 FC F3 D0 EB E7 C0 82 88 " + ;
"EA D0 C9 EC FE E7 C2 B7 B3 82 6E 69 71 53 42 80 " + ;
"64 45 7C 6A 4D B7 A4 97 B3 93 8D 9C 71 62 9C 74 " + ;
"61 BF A4 8A B8 AA 94 34 27 1F D0 C1 BF 76 66 6D " + ;
"FF FD FF B8 B5 AD 8B C0 B3 46 29 14 2B 14 00 75 " + ;
"61 42 E5 DC CE E9 FD EA E8 EB CB A7 7C 55 CE B5 " + ;
"9B FF FE FF 96 8B 8D 05 08 00 EB F7 EB CC B4 C0 " + ;
"48 0F 17 85 7E 4D 94 6C 53 93 71 54 6E 53 38 65 " + ;
"54 47 F0 E7 EA A7 A3 AE FF FD FF FE FB F7 FC FE " + ;
"FE FC FE FE FC FE FE FC FE FE FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FA FF FE FF FF FE " + ;
"FF FC F9 FF FF FC F8 FC F7 F3 FA F7 FB FF FF DA " + ;
"D9 DD B3 CA C2 3A 5D 4F 49 63 55 F3 FF F8 90 A7 " + ;
"99 58 67 5F BF B3 B1 F1 D9 D9 F4 FF F9 64 6A 65 " + ;
"04 03 05 00 00 04 4D 55 55 F3 FB F0 F7 E6 D1 C7 " + ;
"97 7F CA B8 99 E4 C1 A7 D7 A4 90 EC 93 89 FF FF " + ;
"F0 F2 F1 DD EF F0 DC E4 C1 B3 FD DE C9 FF FF FB " + ;
"D9 DE DD 99 90 82 7B 67 4E B7 A8 95 AA 9A 8E 64 " + ;
"49 3F 7F 66 4C 5E 4D 3A E3 DD CA CE CC C1 FF FC " + ;
"FF A4 9C A3 9A 9A 9A EB F0 F1 C6 AE CC 5A 5E 53 " + ;
"AD BC A1 EC FF FF D3 DA E3 4C 0D 0F F5 DC CC ED " + ;
"FF F9 F1 F9 F8 54 53 57 07 20 12 F3 FF FC 83 87 " + ;
"92 3E 05 03 AC 79 65 98 6F 6C 87 62 40 55 3B 17 " + ;
"4A 40 1E FA FC E6 C8 CC C6 FC FE FE FF FC F9 FF " + ;
"FF F8 FC FE FE FC FE FE FC FE FE FC FE FE FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FB FF " + ;
"FF FE FC FB FF FF FC FF FC F7 FF FF FC FB FF FF " + ;
"F8 FD FF F8 FA FF F3 EE EF CC DD D4 A4 B2 B8 B3 " + ;
"BC CA D3 E6 E3 E3 E5 DF 95 74 6B 7A 4A 2E B4 7B " + ;
"65 D2 C3 B0 C8 D9 CE A7 BB BC 00 03 0C 00 01 09 " + ;
"1F 1E 20 7C 78 77 C1 C1 B5 F7 FF F4 EA E5 D6 B0 " + ;
"A0 8F EA D8 C1 D2 BB A1 F6 DC BE F3 F3 CF EA E1 " + ;
"BC F0 FE FA DA EE FF E3 DB DB FF F6 DC F2 D6 B8 " + ;
"FF FC EF DA E7 E9 E3 D3 C6 D3 CB BE A7 A9 95 FC " + ;
"FF F5 FF FA FF 68 60 71 79 7D 7E F3 FE FB C2 B0 " + ;
"A5 98 98 8A FF FF F7 CC E3 DF D0 FF FE FF FB FF " + ;
"B5 A3 B0 34 36 48 3F 63 63 C1 BC B3 D1 C2 C0 F2 " + ;
"F0 FF 7D 74 8E D0 CD C9 4A 39 24 8E 4D 4C 3A 2C " + ;
"10 91 88 6D FF FE E9 C9 C9 BD FF FF FF FD FC FF " + ;
"F9 FB FB FC FF FB FC FE FE FC FE FE FC FE FE FC " + ;
"FE FE FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FD FF FF FF FD FD FF FC FB FF FD FC FF FF " + ;
"FE FD FF FF F7 FB FC FB FE FF FB FF FF FF F8 FA " + ;
"F8 FE FF 45 5F 6D CC CB DF 57 46 4F 60 44 33 C4 " + ;
"8B 64 A2 7E 56 93 60 40 96 61 4C C4 AE A2 E6 F8 " + ;
"F1 A7 BB BC 18 1E 23 00 03 08 04 05 00 0B 17 11 " + ;
"A2 9A 9B CE E0 DF F0 FF F7 E9 F6 E0 BA AA 86 B0 " + ;
"B3 82 D6 CA AE DA D4 D9 E8 EF FF DE E4 DF EE D9 " + ;
"BA C4 94 7C EF CB CB EA E3 F2 B6 A6 9F FC F3 EA " + ;
"FC F7 F6 EB E4 E9 15 0A 0C 09 00 00 06 00 01 07 " + ;
"0A 01 EB E2 B6 3D 46 2C DF FF F6 FF FF F2 E2 CC " + ;
"C0 20 20 1A 23 2E 2B C8 E3 D9 FF FD FF 77 52 5A " + ;
"35 0C 09 E6 E4 CC BD B2 B4 77 7D 90 B1 B6 B4 36 " + ;
"2D 1F AC AA A0 FF FF F7 B4 B5 B3 FE FA FF FF F7 " + ;
"FF FD F6 FF F9 FB FF F3 FD FF FC FE FE FC FE FE " + ;
"FC FE FE FC FE FE FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF F8 F8 F8 FF FF FF FD FF FF FD " + ;
"FF FF FB FA FC FF FF FF FF FF FF FF FE FD F8 FF " + ;
"FF EF FF FF F5 FA F9 FA FE FF 5F 72 81 C1 C0 CA " + ;
"B5 9C 92 3C 32 1A A6 7A 69 90 78 5A B0 8A 68 82 " + ;
"46 30 79 5C 55 D8 E6 E5 EC FF FF BE C9 C7 67 58 " + ;
"66 08 0F 12 02 01 00 0A 06 00 00 07 00 5E 74 62 " + ;
"C8 BE B7 FB E1 E1 F2 EE D5 D4 CB C8 F1 FD FF D0 " + ;
"FF F6 DA FF F5 FB FF FB FF FC F9 FE FF F7 F8 FB " + ;
"EC E7 EF DE FB FB FF D0 CD DC 78 7B 62 6A 6C 4F " + ;
"1E 1F 15 00 09 00 93 89 7F FF FF FF 7A 83 87 36 " + ;
"22 27 12 00 05 A2 B3 B6 FF FF F9 98 82 69 70 34 " + ;
"28 78 56 49 8E 71 62 60 34 1D AA 8B 8A B6 BA BB " + ;
"6D 8D 75 F0 FF F9 EA E7 D8 B9 BD B2 F3 FB F4 FA " + ;
"FF FE FE FE FE FF FF FE FF FF FB FA FF F8 FC FE " + ;
"FE FC FE FE FC FE FE FC FE FE FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FD FF FF " + ;
"F7 FA FE FA FE FF FD FE FF FF FF FF FF FF FC FF " + ;
"FF FB F9 FE FF FF FD FE FF FE FD FF FF FE FF FC " + ;
"FE DD EF EE D6 F3 F0 82 90 8F 86 79 87 A8 98 8C " + ;
"79 4F 32 80 4C 35 C0 A5 A1 FA F6 FF 8B 94 98 8E " + ;
"A5 9D AC 9A 93 C8 B7 B4 C9 C3 C4 CC BD C5 97 B4 " + ;
"B8 35 38 40 00 00 05 00 00 03 00 06 02 13 00 00 " + ;
"30 08 03 3B 21 15 46 39 37 5C 47 56 72 57 60 88 " + ;
"75 66 6E 66 5F 6D 6D 55 79 71 78 38 27 34 75 6C " + ;
"4A FF F9 D7 8C 7C 6F 64 63 3D 3A 32 32 AD B9 BD " + ;
"AB 80 8B 3A 2F 31 E5 FF FF 96 8C A3 6A 31 34 79 " + ;
"4C 21 87 6E 64 A7 7A 54 C8 79 58 78 3B 19 42 3E " + ;
"22 F8 FF FE C3 C9 BE C4 C8 BC E5 DF DA FD FD FD " + ;
"FA FF FF FA FD FF FF FB FF FF FE FF FF FD FC FE " + ;
"FF FB FC FE FE FC FE FE FC FE FE FC FE FE FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FE FE FE FE FE FE FE FE FE FE " + ;
"FE FE FE FE FE FE FE FE FE FE FE FE FE FE 6F 95 " + ;
"95 7B 90 91 C0 C5 C6 51 55 56 AF B8 B5 BD C4 C1 " + ;
"21 1A 17 F4 DD DB 80 4F 47 83 5E 4A 7A 5D 42 9B " + ;
"7D 64 C3 A6 98 F2 E7 E3 CD E9 E3 BF FB F0 8D C2 " + ;
"B5 64 8C 81 48 67 5E 37 55 4A 1C 31 28 1C 24 1D " + ;
"00 05 00 04 11 09 00 1F 1C 0D 2B 26 0D 2B 26 3A " + ;
"55 51 89 9A 96 FF FF F1 E5 D0 B4 95 72 4A 8B 66 " + ;
"58 79 63 40 FF F3 F8 A0 A0 8E FF FF F2 5E 50 51 " + ;
"64 4B 23 B1 7D 6C A8 81 61 57 46 25 25 2B 0C 8B " + ;
"96 7C EF ED DB D5 C6 BD D0 C2 BC FF FF FB FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FE FE FE FE FE FE " + ;
"FE FE FE FE FE FE FE FE FE FE FE FE FE FE FE FE " + ;
"FE FE FB FC FF DD D9 DE 9F A1 A2 CC DA D8 8C A7 " + ;
"A3 70 87 82 F6 FA F5 87 75 74 7B 36 2D 96 5E 4D " + ;
"97 69 51 AB 7E 63 9E 6D 57 4D 27 15 BA AE 9C E0 " + ;
"EC D6 D9 DA D8 FF FD FC A5 A0 9D F3 ED E8 FF FF " + ;
"F7 F5 FD F2 F6 FF F7 E6 F9 E8 FF F8 FC FD FC FF " + ;
"F8 FF FF E0 DE DE C5 AC AA A3 78 6F 83 52 42 99 " + ;
"6A 55 B6 8A 6B 6A 3D 32 82 68 50 EA D9 D6 18 11 " + ;
"0E DA D3 BF 71 54 50 79 54 40 47 2E 1A 99 92 7E " + ;
"EC F8 E2 F2 FD ED D7 D6 CC F0 E4 E2 FF FE FE FB " + ;
"FB FB FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FE FE " + ;
"FE FE FE FE FE FE FE FE FE FE FE FE FE FE FE FE " + ;
"FE FE FE FE FE FE FA FF FF FF F9 FE FF FC FF F7 " + ;
"F1 F2 CC D4 D3 AA BE B9 90 A2 9B 48 54 4E AB A2 " + ;
"99 BB B1 A7 61 4E 41 53 31 21 75 48 33 A0 79 63 " + ;
"7B 6A 55 BC BD A9 A8 9C 90 61 56 48 BE A4 96 65 " + ;
"39 28 81 4F 3B 83 5B 42 83 61 43 83 5C 3C 61 3E " + ;
"30 FF FE F5 5D 68 66 65 63 63 5D 3A 36 A1 6A 5D " + ;
"B2 7E 6D A9 84 6E BB 84 5F BF 84 82 5D 3A 18 A8 " + ;
"93 92 03 00 02 A4 B5 9A D2 D4 DF B9 BF AC E1 D9 " + ;
"D2 F4 F9 F0 E1 F0 E8 E0 EC E6 FE FE FE FF FE FF " + ;
"FB FC FF F3 FC FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FE FE FE FE FE FE FE FE FE FE FE FE FE FE " + ;
"FE FE FE FE FE FE FE FE FE FE F2 FC FC F4 F6 F7 " + ;
"FF FD FF FF FF FF FA FF FF F6 FF FE F4 FB F8 BF " + ;
"C0 BE 7E 9A 9B 74 8E 94 91 A7 AD 9B A7 A7 AB AB " + ;
"A5 55 53 49 95 99 94 F6 FF FF 48 56 4A C0 C4 B8 " + ;
"F2 E1 D4 7D 50 42 B0 73 5F B9 79 60 B5 75 57 C2 " + ;
"7F 60 A3 76 55 74 5F 49 BC C2 B7 C8 C8 C8 5E 42 " + ;
"41 B7 86 7C 90 60 48 7B 55 33 40 46 27 52 53 39 " + ;
"77 74 66 28 27 1D C2 C6 C1 FB FF FF E6 E3 DE E5 " + ;
"DA D6 BA BF BE EE F9 F7 F4 FF FF F7 FC FD FF FE " + ;
"FF FD F8 FA F8 FB FF F4 FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FB FF " + ;
"FF FB FF FF F7 FC FB F3 F9 F8 F6 FF FF FA FF FF " + ;
"FF FE FF FF FC FC F8 FE FF F0 F8 FF A1 A8 B9 87 " + ;
"8D 9A 71 77 7E 90 99 9D C3 D2 DB 87 9B AC 6F 7C " + ;
"8A F8 F9 FF 6C 62 6E 4B 3B 42 73 5E 60 7D 69 64 " + ;
"6E 5F 56 65 5F 52 84 75 5B 41 3C 27 ED F3 E8 F0 " + ;
"F8 F8 92 90 96 58 4C 4C 8E 7B 6C AA 95 7A 98 AC " + ;
"9F E7 FB DE F5 F5 FB AC B2 A7 C8 D0 C6 AD AC B6 " + ;
"DB DD C9 FF F7 F7 FA FF FF F6 FC FB FA FF FF FD " + ;
"FE FC FE F9 F8 FF FE FD FF FF FE F6 FD FA FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF 00 " + ;
"00 00 FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF F6 FE FD F8 FD FC FF FD FE FF FE FF FF FC " + ;
"FD FB FB FB FB FF FF F9 FF FE FF FE FA FF FE FF " + ;
"FF FC FF FF FB FF FF FD FF DD DB E1 A3 AC B5 A3 " + ;
"B1 BD B1 B4 C3 94 8C 9D 8D 84 91 73 74 7E 5C 62 " + ;
"67 6B 70 71 4D 5A 58 5B 77 71 63 77 78 7E 8A 8A " + ;
"99 A1 A0 6E 77 7B 8B 99 A5 8F 9F AB BD C7 CE CF " + ;
"D1 D1 EF DD DE E5 D2 CF E0 CA CF F1 E5 E5 F3 F7 " + ;
"F2 F6 FF FE F6 FF F7 F6 FF F7 FA FE F9 FF FF FC " + ;
"FF FF FC FF FD F9 FF FF FC FF FF FC FF FF FB FF " + ;
"FF FC FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF 00 00 00 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FC FF FD FF FF FF FF FF FF FF " + ;
"FE FF FF FD FE FE FE FE FB FF FF F8 FF FF FF FF " + ;
"FB FF FF F8 FD FF F9 F7 FE F9 FA FF FF FA FC FD " + ;
"FF FC FE FF FF FE FF FF FE FE FC FB FF FF FE FD " + ;
"FE FC E5 E6 E4 E7 E5 E4 BB B7 B6 CA C6 C5 E8 F1 " + ;
"FA E0 E2 EA FC F4 FB EE E3 EB F2 ED F6 F8 F7 FF " + ;
"FB FC FF FF FB FF FD FF FE FE FA FF FF FF FC FD " + ;
"F7 FC FF FE FF FE FF F6 FF FC FF FF FF F9 FB FA " + ;
"F6 FF FF FB FF FF FC FF FF FE FF FF FE FC FD FB " + ;
"FF FF FE FA F8 F8 FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF 00 00 00 FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FB FF FE FB FF FE " + ;
"FF FF FF FF FF FF FF FE FF FF FE FF FF FC FE FB " + ;
"FD FE FB FD FE FF FF FC FF FF FA FF FF FE F6 F5 " + ;
"F9 FD FD FF FC FE FE FC FF F6 FF FE FB FD FF FE " + ;
"F8 FF FC FE FE FE FF FE FF FD FD FF FF FA FF FF " + ;
"FA FF F7 FF F5 FD FF FE F9 FB FC FF FE FF FF FE " + ;
"FA FF FF FB FE FF FB F7 FC FB F3 FE FB F9 FF FE " + ;
"F6 FB F9 F8 FF FF F2 FC FC F8 FF FF FF FE FF FF " + ;
"FB FD FF FF FE FF FD FC FF FF FF F9 FE FD F7 FF " + ;
"FF F8 FF FF F9 FC FF FF FE FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF FF " + ;
"FF FF FF FF FF FF FF FF FF FF FF 00 00 00 "

RETURN NIL

* ============================= EOF of hexbincnt.prg ==================================
